"""Class for tokenizing assembly lines, as well as other tokenization constants"""

import re
from ..utils import ParameterSaver, eq_obj, hash_obj
from .norm_utils import *
from enum import Enum
from ..utils.type_utils import *


# Token type names
class Tokens:
    INSTRUCTION_ADDRESS = 'inst_addr'
    INSTRUCTION_START = 'inst_start'
    SPLIT_IMMEDIATE = 'split_imm'
    DISASSEMBLER_INFO = 'disassembler_info'
    NEWLINE = 'newline'
    SPACING = 'spacing'

    OPEN_BRACKET = 'open_bracket'
    CLOSE_BRACKET = 'close_bracket'
    PLUS_SIGN = 'plus_sign'
    TIMES_SIGN = 'times_sign'
    COLON = 'colon'

    INSTRUCTION_PREFIX = 'prefix'
    OPCODE = 'opcode'
    REGISTER = 'register'
    IMMEDIATE = 'immediate'

    MEMORY_SIZE = 'memory_size'
    MEMORY_EXPRESSION = 'memory_expression'
    BRANCH_PREDICTION = 'branch_prediction'
    STRING_LITERAL = 'string_literal'
    SEGMENT_ADDRESS = 'segment_address'

    MISMATCH = 'mismatch'


class TokenizationLevel(Enum):
    """Different levels to perform tokenization"""
    OPCODE = ['op', 'opcode', 'operand', 'opcodes', 'operands']
    INSTRUCTION = ['inst', 'instruction', 'line', 'instructions', 'lines']
    AUTO = ['auto', 'automatic', 'default']


class Architectures(Enum):
    """Known (but not necessarily supported) architectures"""
    X86 = ['x86', 'i686', 'x86_64']
    JAVA = ['java', 'java_bytecode']


def get_architecture(arch: 'Union[str, Architectures]') -> 'Architectures':
    """Returns the architecture
    
    Args:
        arch (Union[str, Architectures])
    """
    if isinstance(arch, Architectures):
        return arch
    elif isinstance(arch, str):
        arch = arch.lower().replace('-', '_')
        for a in Architectures:
            if any(arch == v for v in a.value):
                return a
        raise ValueError("Unknown architecture string: %s" % repr(arch))
    else:
        raise TypeError("Cannot get architecture from object of type: %s" % repr(type(arch).__name__))


class TokenMismatchError(Exception):
    pass

class UnknownTokenError(Exception):
    pass


# Special tokens to insert into tokenizer by default
SPECIAL_TOKENS_START = [
    (Tokens.STRING_LITERAL, RE_STRING_LITERAL),
    (Tokens.DISASSEMBLER_INFO, RE_DISASSEMBLER_INFO),
    (Tokens.INSTRUCTION_START, INSTRUCTION_START_TOKEN),
    (Tokens.SPLIT_IMMEDIATE, SPLIT_IMMEDIATE_TOKEN),
    (Tokens.PLUS_SIGN, RE_PLUS_SIGN),
    (Tokens.TIMES_SIGN, RE_TIMES_SIGN),
    (Tokens.OPEN_BRACKET, RE_OPEN_BRACKET),
    (Tokens.CLOSE_BRACKET, RE_CLOSE_BRACKET),
    (Tokens.COLON, RE_COLON),
    (Tokens.SPACING, RE_SPACING),
    (Tokens.NEWLINE, RE_NEWLINE),
    (Tokens.IMMEDIATE, RE_IMMEDIATE),
]
SPECIAL_TOKENS_END = [
    (Tokens.MISMATCH, r'.'),
]


# An indication that the default newline tuple should be used when tokenizing/normalizing
_USE_DEFAULT_NT = object()


def parse_tokenization_level(tokenization_level, auto_tl):
    """Returns the bincfg.TokenizationLevel enum based on the given tokenization_level.

    Args:
        tokenization_level (Union[bincfg.TokenizationLevel, str]): either a string tokenization level, or a class from the 
            bincfg.TokenizationLevels enum
        auto_tl (bincfg.TokenizationLevel): the default tokenization level to use if we get an 'auto' tokenization level

    Returns:
        bincfg.TokenizationLevel: a class from the ``bincfg.TokenizationLevels`` enum
    """
    if not isinstance(auto_tl, TokenizationLevel) or auto_tl == TokenizationLevel.AUTO:
        raise TypeError("`auto_tl` must be a bincfg.TokenizationLevel, and cannot be bincfg.TokenizationLevel.AUTO. Got: %s" % repr(auto_tl))
    
    if isinstance(tokenization_level, str):
        tl = tokenization_level.lower().replace('-', '_')
        for l in TokenizationLevel:
            if tl in l.value:
                ret = l
                break
        else:
            raise ValueError("Unknown tokenization_level string: '%s'" % tokenization_level)
    elif isinstance(tokenization_level, TokenizationLevel):
        ret = tokenization_level
    else:
        raise TypeError("Unknown tokenization_level type: '%s'" % type(tokenization_level))
    
    # Check for auto
    if ret is TokenizationLevel.AUTO:
        return auto_tl
    
    return ret


class BaseTokenizer(metaclass=ParameterSaver):
    """A default class to tokenize instructions

    Should be subclassed once for each instruction set, providing the tokens being used.

    Many functions may be overriden to change tokenization behavior. These functions all start with the name `token\_...`
    and take as input a single state dictionary and return either a string for the next token to append to the current
    line being tokenized, or None to not add anything to the line. The state dictionary contains the following:

        - 'tokenizer' (BaseTokenizer): this tokenizer 
        - 'kwargs' (Dict[str, Any]): dictionary of extra kwargs passed to the initial call to the `tokenize` function
        - 'all_strings' (List[str]): list of input strings (args) passed to the initial call to the `tokenize` function
        - 'token_handlers' (Dict[str, Callable[]]): dictionary mapping token types to the function that handles that token
        - 'sentence' (List[Tuple[str, str]]): list of processed token tuples to return, each a tuple of (token\_name, token)
        - 'newline_tup' (Union[None, Tuple[str, str]]): token tuple to add at the end of each line to indicate a new line
        - 'match_instruction_address' (bool): whether or not we are matching instruction addresses
        - 'split_imm' (bool): whether or not we are currently handling an immediate token that was split 
        - 'line' (List[Tuple[str, str]]): the current line of tokens we are working on
        - 'string' (str): the current string being tokenized
        - 'token_type' (str): the type of the 'token', should be from `bincfg.normalization.base_tokenizer.Tokens`
        - 'token' (str): the currently matched token string
        - 'match' (re.Match): the re match object that matched this token
    
    Some extra functions are available for overriding including:

        - handle_line(): called at the end of each line being tokenized (an individual string passed to the tokenizer)
        - handle_sentence(): called at the end of each sentence being tokenized (aggregation of all lines passed to the tokenizer)
    
    Each instruction set architecture (ISA) should have its own ``Tokenizer`` class that inherits from ``BaseTokenizer``. The 
    tokenization process uses python's ``re`` module to perform tokenization, converting strings into streams of 
    (token\_name, token\_string) tuples. For more information on how to use regex to create tokenizers, see: 
    https://docs.python.org/3/library/re.html#writing-a-tokenizer
    
    TOKENIZATION PROCESS
    
        1. Clean the incomming instruction strings using the passed `clean_instruction_func`
        2. Iterate through the strings finding all tokens
        
           a. Each token is sent to its corresponding token handler function
           b. At the end of each 'line' (EG: end of a passed `string`, reaching Tokens.NEWLINE token, etc.), that line
              is handled with the `handle_line()` function
           c. All tokens are added to the same return 'sentence', even if multiple strings in `strings` were passed

        3. After all strings have been tokenized and lines handled, the final return 'sentence' is sent to `handle_sentence()`
    
    SPECIAL TOKENS

    There are some 'special tokens' that are assumed to exist for all ISA's as they are a part of the tokenization
    process itself. These tokens will be inserted into the passed `tokens` parameter at the beginning of the list
    (IE: they are the first tokens searched for), except for the 'mismatch' token which is inserted at the end,
    and are inserted in the following order:

        1. String literals (Tokens.STRING_LITERAL) - matches strings which can start/end with matching single or double 
           quotes, and can escape inner quotes with \\' or \\", and can escape the escape character with \\\\. Any extra
           escape characters (not behind a ' or " or \\) will be left as-is.
        2. Disassembler information (Tokens.DISASSEMBLER_INFO) - matches disassembler information of the form "<...>". 
           This info must be within open/close angle brackets. It is also possible to nest angle brackets within the 
           disassembler info up to a maximum current depth of 3. IE: we can match the following:
        
            * "<no angle brackets inside>" - depth of 1
            * "<angle <brackets> depth <2>>" - depth of 2
            * "<level <3 angle <bracket>> depth>" - depth of 3
        
           We also do not check that every open has a matching close, just that every close has a matching open. So, the
           following could still be matched:

            * "<lots of <<<<<<< things>"
        
           However, missing or unmatched ending angle brackets will fail, as well as very deep nesting:

            - "<" : no matching '>' only for the first occurance of '<'
            - "<data>>" : no matching '<' for both of the '>' brackets
            - "<super<deep<nested<...<thing>>...>>" : too large nesting depth

           String literals are checked first within the disassembler info so that any end brackets '>' within the strings
           won't affect the parsing of the disassembler info.
        
           This limit on nesting depth is present due to the inability for python's re engine to handle recursive matching
           of nested brackets, and I can't think of any way to implement it entirely within re's (which is needed in order
           to continue using the python re tokenization method). I don't see any reason why this would be needed as we
           already go down to a depth of 3 to handle more than what I would expect as output from disassemblers, and if
           the user is inserting information themselves, they could simply input the information within the brackets
           using a different delimiter and parse it themselves by overriding things like `token_disassembler_info()`
           and `handle_disassembler_info` in the `Tokenizer` and `Normalizer` classes respectively. If a larger
           depth is needed, one can manually alter the `_DIS_INFO_MAX_REC_DEPTH` variable at the top of this file. It will
           increase the valid nesting depth at the cost of slower regular expression matching for disassembler info.
        3. Instruction start token "#start_instr#" (Tokens.INSTRUCTION_START) - used to determine when instructions 
           start/stop when using an op-level tokenization scheme. When tokenizing, we need to know when a new instruction 
           is started to decide if an immediate value found should be considered an instruction address or just a plain 
           immediate. New instructions occur whenever we reach a newline token, an instruction start token, or the start 
           of a new string passed in the args of the `tokenize()` method. This instruction start token is removed when 
           found, and won't appear during normalization.
        4. Split immediate token "#split_imm#" (Tokens.SPLIT_IMMEDIATE) - used to designate a split immediate value. This 
           is useful for reducing the number of unique tokens present while keeping full immediate information. When using 
           split immediates during normalization, immediate values with more digits than some threshold will be split into multiple
           immediate tokens and placed one after the other, prepended with this "#split_imm#" token. In order to keep
           that output as renormalizable, the tokenizer, when finding one of these split immediate tokens, will concatenate
           all of the following immediate tokens until reaching some non-immediate (and, non-spacing) token to rebuild
           the original immediate token. This split immediate token is removed when found, and won't appear during normalization
        5. Plus sign (Tokens.PLUS_SIGN) - '+'
        6. Times sign (Tokens.TIMES_SIGN) - '*'
        7. Open bracket (Tokens.OPEN_BRACKET) - '['
        8. Close bracket (Tokens.CLOSE_BRACKET) - ']'
        9. Colon (Tokens.COLON) - ':'
        10. Spacing (Tokens.SPACING) - One or more space ' ', comma ',', or tab '\\t' characters in a row
        11. Newline (Tokens.NEWLINE) - Either the newline character '\\n' or a pipe character '|'
        12. Immediate values (Tokens.IMMEDIATE) - any integer immediate value in hex, decimal, octal, or binary. Hex values
            must start with '0x', octal with '0o', and binary with '0b'
        13. Mismatch token (Tokens.MISMATCH) - matches any character. Inserted at the very end of `tokens` and is used 
            to designate the start of an unknown token or character so that can be handled (by default, an error is raised)

    If you wish to keep some of the above tokens, but overwrite others, you can set that token's regex in the passed
    `tokens` parameter, and that will overwrite these special tokens. You may also set it to None to not insert it at all.

    INSTRUCTION ADDRESSES

    If match_instruction_address=True when tokenizing, the tokenizer will attempt to match instruction addresses at the
    beginning of each line. If there is an immediate value at the start of a line (IE: start of a string in `strings`, 
    or immediately after a Tokens.NEWLINE or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token
    will be converted into a Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token 
    (again, ignoring any Tokens.SPACING), then that first Tokens.COLON match will be appended to that Tokens.INSTRUCTION_ADDRESS 
    token, removing any Tokens.SPACING inbetween them. For example, using the x86 tokenization scheme:

        - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
        - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
        - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
    
           
    Parameters
    ----------
    tokens: `Optional[List[Tuple[str, str]]]`
        the tokens to use. Should be a list of 2-tuples. Each tuple is a pair of (name, regex) where
        name is the string name of the token, and regex is a regular expression to find that token. These
        tuples should be ordered in the preferred order to search for tokens. If None, then this will default to 
        self.DEFAULT_TOKENS (which should be set when defining the class)
    token_handlers: `Optional[Dict[str, Callable[[Dict[str, Any]], Union[None, str]]]]`
        optional dictionary mapping token type strings to functions to handle those token types when tokenizing. This is
        intended to be used when you wish to add entirely new token types not present in `bincfg.normalization.base_tokenizer.Tokens`.
        If you wish to change the behavior of handling an already-present token type, just override that token handler function.
        These will override the default token handlers.
    insert_special_tokens: `bool`
        by default, some special tokens will be inserted at the front of `tokens` (see the 'special tokens' listed above).
        If you wish to stop this from happening, you can set `insert_special_tokens` to False
    case_sensitive: `bool`
        If True, then regular expressions will be matched exactly as they appear. If False, then the re.IGNORECASE flag
        will be passed when compiling the regular expressions
    """

    DEFAULT_NEWLINE_TUPLE = (Tokens.NEWLINE, '\n')
    """The default (token_type, token) tuple to use for newlines"""

    ARCHITECTURE = None
    """The architecture this tokenizer works on"""

    def __init__(self, tokens=None, token_handlers=None, insert_special_tokens=True, case_sensitive=False):
        self.tokens = tokens if tokens is not None else self.DEFAULT_TOKENS
        self.case_sensitive = case_sensitive

        # Insert the special tokens, make sure user values override regex's and positions of default tokens, and remove
        #   any tokens that the user has set to None
        if insert_special_tokens:
            user_tokens = set(t[0] for t in self.tokens)
            self.tokens = [t for t in SPECIAL_TOKENS_START if t[0] not in user_tokens] + self.tokens + \
                [t for t in SPECIAL_TOKENS_END if t[0] not in user_tokens]
            self.tokens = [t for t in self.tokens if t[1] is not None]

        self.token_handlers = token_handlers if token_handlers is not None else {}
        self._init_tokenizer()
    
    def _init_tokenizer(self):
        """Initializes the tokenizer from self.tokens"""
        flags = (re.M|re.UNICODE)
        flags = (flags|re.IGNORECASE) if not self.case_sensitive else flags
        self.tokenizer = re.compile('|'.join([("(?P<%s>%s)" % pair) for pair in self.tokens]), flags=flags)
    
    def tokenize(self, *strings, newline_tup=_USE_DEFAULT_NT, match_instruction_address=True, **kwargs):
        """Tokenizes the input
        
        Subclasses should override any self.token_* methods they wish to inject behavior into. Each one of those functions
        takes in a 'state' dictionary as input and should return either a new string token or None to use the old token.

        See the docs for :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info on how tokenization
        works, how to create subclasses, etc.

        Args:
            strings (str): arbitrary number of strings to tokenize.
            newline_tup (Optional[Tuple[str, str]]): the tuple to insert inbetween each passed string, or None to not 
                insert anything. Defaults to `self.__class__.DEFAULT_NEWLINE_TUPLE`.
            match_instruction_address (bool, optional): if True, will match instruction addresses. If there is an immediate
                value at the start of a line (IE: start of a string in `strings`, or immediately after a Tokens.NEWLINE
                or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token will be converted into a
                Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token (again, ignoring
                any Tokens.SPACING), then that first Tokens.COLON match will be appended to that Tokens.INSTRUCTION_ADDRESS 
                token, removing any Tokens.SPACING inbetween them. For example, using the x86 tokenization scheme:

                    - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
                
            kwargs (Any): extra kwargs to store in the tokenizer state, for use in child classes

        Returns:
            List[Tuple[str, str]]: list of (token_type, token) tuples
        """
        if len(strings) == 0:
            return []
        
        # Replace the newline_tup if using
        newline_tup = self.DEFAULT_NEWLINE_TUPLE if newline_tup is _USE_DEFAULT_NT else newline_tup
        
        # If the first element starts with an `INSTRUCTION_START` token, then concatenate all of the strings together
        if strings[0].strip().startswith(INSTRUCTION_START_TOKEN):
            strings = [' '.join(strings)]
        
        # Dictionary mapping token types to self functions to handle those tokens. Only have to add tokens to handle
        token_handler_dict = {
            Tokens.INSTRUCTION_ADDRESS: self.token_instruction_address,
            #Tokens.INSTRUCTION_START: None,
            #Tokens.SPLIT_IMMEDIATE: None,
            Tokens.DISASSEMBLER_INFO: self.token_disassembler_info,
            Tokens.NEWLINE: self.token_newline,
            Tokens.SPACING: self.token_spacing,

            Tokens.OPEN_BRACKET: self.token_all_symbols,
            Tokens.CLOSE_BRACKET: self.token_all_symbols,
            Tokens.PLUS_SIGN: self.token_all_symbols,
            Tokens.TIMES_SIGN: self.token_all_symbols,
            Tokens.COLON: self.token_all_symbols,

            Tokens.INSTRUCTION_PREFIX: self.token_instruction_prefix,
            Tokens.OPCODE: self.token_opcode,
            Tokens.REGISTER: self.token_register,
            Tokens.IMMEDIATE: self.token_immediate,

            Tokens.MEMORY_SIZE: self.token_memory_size,
            #Tokens.MEMORY_EXPRESSION: 'memory_expression'
            Tokens.BRANCH_PREDICTION: self.token_branch_prediction,
            Tokens.STRING_LITERAL: self.token_string_literal,

            Tokens.MISMATCH: self.token_mismatch,
        }
        token_handler_dict.update(self.token_handlers)
        
        token_state = {'kwargs': kwargs, 'all_strings': strings, 'token_handlers': token_handler_dict, 'sentence': [], 
                       'newline_tup': newline_tup, 'match_instruction_address': match_instruction_address, 'tokenizer': self}
        
        for string in strings:
            token_state.update({'previous_newline': True, 'split_imm': False, 'line': [], 'string': string, 'matched_ia_colon': False})

            for mo in self.tokenizer.finditer(string):
                token_state['token_type'], token_state['token'], token_state['match'] = mo.lastgroup, mo.group(), mo
                
                # Handle all the various tokens. Keep all tokens
                new_token = token_state['token_handlers'][token_state['token_type']](token_state) if token_state['token_type'] in token_state['token_handlers'] \
                    else token_state['token'] if token_state['token_type'] in [Tokens.SPLIT_IMMEDIATE, Tokens.INSTRUCTION_START] \
                    else self.token_unknown(token_state)
                        
                if new_token is not None:
                    token_state['line'].append((token_state['token_type'], new_token))
            
            # If there are split immediates, parse them out now in line
            token_state['line'] = self._merge_split_immediates(token_state['line'])

            # Check for instruction addresses
            if token_state['match_instruction_address']:
                token_state['line'] = self._check_instruction_address(token_state['line'])

            # Handle the new line and add on the newline_tup if using, and we don't have an empty sentence
            new_line = self.handle_line(token_state)
            token_state['sentence'] += new_line if new_line is not None else token_state['line']
            if token_state['newline_tup'] is not None and len(token_state['sentence']) > 0:
                token_state['sentence'].append(token_state['newline_tup'])
        
        return self.handle_sentence(token_state)
    
    def _merge_split_immediates(self, line):
        """Merges any split immediates into one, ignoring any spacing inbetween them"""
        ret_line, idx = [], 0
        while idx < len(line):
            # If this isn't a split immediate, just go ahead to the next token
            if line[idx][0] not in [Tokens.SPLIT_IMMEDIATE]:
                ret_line.append(line[idx])
                idx += 1
                continue

            # This is a split immediate token, scan ahead for all immediate tokens (including instruction address), 
            #   ignoring spacing, until reaching a non-immediate token. Concatenate these into a single immediate value
            imm_inds = scan_for_token(line, type=Tokens.IMMEDIATE, ignore_type=Tokens.SPACING, stop_unmatched=True,
                                      start=idx + 1, max_matches=None, wrap=False, on_no_match=[])
            if len(imm_inds) > 0:
                ret_line.append((line[imm_inds[0]][0], ''.join([line[i][1] for i in imm_inds])))
            idx = max(imm_inds + [idx]) + 1
        
        return ret_line
    
    def _check_instruction_address(self, line):
        """Scans through the current line checking for possible instruction address locations"""
        _newlines = [Tokens.NEWLINE, Tokens.INSTRUCTION_START]
        ret_line, idx = [], 0
        while idx < len(line):
            token_tup, token_type = line[idx], line[idx][0]

            # If we are at the start of a line, check for instruction address
            if idx == 0 or token_type in _newlines:
                inst_addr_inds = scan_for_token(line, type=Tokens.IMMEDIATE, ignore_type=Tokens.SPACING, stop_unmatched=True, 
                                                start=(idx+1) if token_type in _newlines else idx, wrap=False, on_no_match=[], ret_list=True)
                
                if len(inst_addr_inds) > 0 and imm_to_int(line[inst_addr_inds[0]][1]) > 0:
                    # Check for possible colon token
                    inst_addr_inds += scan_for_token(line, type=Tokens.COLON, ignore_type=Tokens.SPACING, stop_unmatched=True, 
                                                     start=max(inst_addr_inds)+1, wrap=False, on_no_match=[], ret_list=True)
                    
                    # Build the instruction address token. Insert any tokens before the start if present
                    for i in range(idx if token_type in _newlines + [Tokens.SPACING] else (idx + 1), min(inst_addr_inds)):
                        ret_line.append(line[i])
                    token_tup = (Tokens.INSTRUCTION_ADDRESS, ''.join([line[i][1] for i in inst_addr_inds]))
                    idx = max(inst_addr_inds)

            # Add in this token and increment the index
            ret_line.append(token_tup)
            idx += 1

        return ret_line
    
    def handle_line(self, state):
        """Handles a single line (one string passed to the tokenizer)
        
        Each line could contain newlines and whatnot, but no newline_tup's will have been inserted.

        Subclasses may override this function for more behavior, but it defaults to just returning the passed line.

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            List[Tuple[str, str]]: list of (token_type, token) tuples for this line
        """
        return state['line']
    
    def handle_sentence(self, state):
        """Handles an entire sentence (aggregation of all strings passed to one call of this tokenizer)
        
        Inbetween each line, a newline_tup will have already been inserted (if using)

        Subclasses may override this function for more behavior, but it defaults to just returning the passed sentence

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            List[Tuple[str, str]]: the final list of tokens
        """
        return state['sentence']

    def token_branch_prediction(self, state):
        """Handles any branch_prediction tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']

    def token_immediate(self, state):
        """Handles any immediate tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']

    def token_instruction_prefix(self, state):
        """Handles any instruction_prefix tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']

    def token_memory_size(self, state):
        """Handles any memory_size tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']

    def token_opcode(self, state):
        """Handles any opcode tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']

    def token_register(self, state):
        """Handles any register tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']

    def token_disassembler_info(self, state):
        """Handles any disassembler information tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']
    
    def token_string_literal(self, state):
        """Handles any string literals
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']
    
    def token_spacing(self, state):
        """Handles any spacing tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']
    
    def token_newline(self, state):
        """Handles any newline tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']
    
    def token_instruction_address(self, state):
        """Handles any instruction address tokens
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']
    
    def token_all_symbols(self, state):
        """Handles all symbol tokens ('+', '*', '[', ']', ':')
        
        This can be overriden by subclasses for more functionality, but defaults to just returning the original token, 
        except for colons ':', for which we check if the previous non-spacing token was an immediate value. If so, and
        `match_instruction_address` is True, then we append any inbetween spacing and the colon to that immediate and 
        replace its type with Token.INSTRUCTION_ADDRESS.

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info
        
        Returns:
            Union[str, None]: either a string token for the next token to append to line, or None to not append anything
        """
        return state['token']
    
    def token_mismatch(self, state):
        """What to do when there is a token mismatch in a string
        
        This can be overriden by subclasses for more functionality, bet defaults to raising a ``TokenMismatchError`` with 
        info on the mismatch

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info

        Raises:
            TokenMismatchError: by default
        """
        raise TokenMismatchError("Mismatched token '%s' at index %d in string: %s" % (state['token'], state['match'].start(), repr(state['string'])))
    
    def token_unknown(self, state):
        """What to do when there is a token type that we don't know how to handle
        
        This can be overriden by subclasses for more functionality, bet defaults to raising a ``UnknownTokenError`` with 
        info on the unknown token

        Args:
            state (Dict): dictionary of current state. See :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info

        Raises:
            UnknownTokenError: by default
        """
        raise UnknownTokenError("Unknown token type %s, token: %s" % (repr(state['token_type']), repr(state['token'])))

    def __call__(self, *strings, newline_tup=_USE_DEFAULT_NT, match_instruction_address=True, **kwargs):
        """Tokenizes the input
        
        Subclasses should override any self.token_* methods they wish to inject behavior into. Each one of those functions
        takes in a 'state' dictionary as input and should return either a new string token or None to use the old token.

        See the docs for :func:`~bincfg.normalization.base_tokenizer.BaseTokenizer` for more info on how tokenization
        works, how to create subclasses, etc.

        Args:
            strings (str): arbitrary number of strings to tokenize.
            newline_tup (Optional[Tuple[str, str]]): the tuple to insert inbetween each passed string, or None to not 
                insert anything. Defaults to `self.__class__.DEFAULT_NEWLINE_TUPLE`.
            match_instruction_address (bool, optional): if True, will match instruction addresses. If there is an immediate
                value at the start of a line (IE: start of a string in `strings`, or immediately after a Tokens.NEWLINE
                or Tokens.INSTRUCTION_START [ignoring any Tokens.SPACING]), then that token will be converted into a
                Tokens.INSTRUCTION_ADDRESS token. If there is a Tokens.COLON immediately after that token (again, ignoring
                any Tokens.SPACING), then that first Tokens.COLON match will be appended (along with any inbetween Tokens.SPACING)
                to that Tokens.INSTRUCTION_ADDRESS token. For example, using the x86 tokenization scheme:

                    - "0x1234: add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234:'), ...]
                    - "  0x1234     : add rax rax" -> [(Tokens.SPACING, '  '), (Tokens.INSTRUCTION_ADDRESS, '0x1234     :'), ...]
                    - "0x1234 add rax rax" -> [(Tokens.INSTRUCTION_ADDRESS, '0x1234'), ...]
                
            kwargs (Any): extra kwargs to store in the tokenizer state, for use in child classes

        Returns:
            List[Tuple[str, str]]: list of (token_type, token) tuples
        """
        return self.tokenize(*strings, newline_tup=newline_tup, match_instruction_address=match_instruction_address, **kwargs)
    
    def __repr__(self) -> str:
        if eq_obj(self.tokens, self.DEFAULT_TOKENS):
            tokens_str = ''
        elif len(repr(self.tokens)) > 30:
            tokens_str = '...'
        else:
            tokens_str = 'tokens=' + repr(self.tokens)
        return self.__class__.__name__ + "(%s)" % tokens_str
    
    def __str__(self) -> str:
        return self.__class__.__name__
    
    def __eq__(self, other):
        return type(self) == type(other) and eq_obj(self.tokens, other.tokens) and self.case_sensitive == other.case_sensitive \
            and eq_obj(self.token_handlers, other.token_handlers)
    
    def __hash__(self):
        return hash_obj([type(self).__name__, self.tokens, self.case_sensitive, self.token_handlers], return_int=True)




from ..base_normalizer import BaseNormalizer
from ..base_tokenizer import TokenizationLevel
from .java_tokenizer import JAVA_DEFAULT_TOKENIZER
from ..norm_funcs import threshold_immediate, replace_immediate
from ...utils import hash_obj


# All of these are the 'invoke' commands
JAVA_FUNCTION_CALL_RE = r'invoke(?:virtual|interface|special|static|dynamic)'

# Three types: conditional (the 'if' ones), tabular (switch/lookup tables), and unconditional (goto, jsr, etc)
# This one does technically match to 4 instructions that don't exist: if_acmp[lt,le,gt,ge], but that should be fine, this looks nicer
_JAVA_CONDITIONAL_JUMP_RE = r'if(?:{null}|(?:_[ia]cmp)?{comps})'.format(
    null=r'(?:non)?null',
    comps=r'(?:eq|ne|lt|le|gt|ge)',
)
_JAVA_UNCONDITIONAL_JUMP_RE = r'(?:goto|jsr)(?:_w)?'
_JAVA_SWITCH_JUMP_RE = r'(?:table|lookup)switch'
JAVA_JUMP_RE = r'{cond}|{switch}|{uncond}'.format(cond=_JAVA_CONDITIONAL_JUMP_RE, uncond=_JAVA_UNCONDITIONAL_JUMP_RE, switch=_JAVA_SWITCH_JUMP_RE)


class JavaBaseNormalizer(BaseNormalizer):
    """A base class for a normalization method. 
    
    Performs an 'unnormalized' normalization, removing what is likely extraneous information, and providing a base class
    for other normalization methods to inherit from.
    
    Parameters
    ----------
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """

    DEFAULT_TOKENIZATION_LEVEL = TokenizationLevel.INSTRUCTION
    """The default tokenization level used for this normalizer"""

    renormalizable = True
    """Whether or not this normalization method can be renormalized later by other normalization methods"""

    tokenizer = None
    """The tokenizer used for this normalizer"""

    token_sep = None
    """The separator string used for this normalizer
    
    Will default to ' ' for ``BaseNormalizer``, and '_' for all other normalizers.
    """

    tokenization_level: TokenizationLevel
    """The tokenization level to use for this normalizer"""

    def __init__(self, tokenizer=None, token_handlers=None, token_sep=None, tokenization_level=TokenizationLevel.AUTO, anonymize_tokens=False):
        super().__init__(tokenizer=tokenizer if tokenizer is not None else JAVA_DEFAULT_TOKENIZER, token_handlers=token_handlers, 
                         token_sep=token_sep if token_sep is not None else ' ',
                         tokenization_level=tokenization_level, anonymize_tokens=anonymize_tokens)
        self.register_opcode_handler(JAVA_FUNCTION_CALL_RE, self.opcode_function_call)
        self.register_opcode_handler(JAVA_JUMP_RE, self.opcode_jump)
    
    def opcode_function_call(self, state):
        """Handles function call opcodes, defaults to doing nothing"""
        return None
    
    def opcode_jump(self, state):
        """Handles jump opcodes, defaults to doing nothing"""
        return None


class JavaReplaceImmediateNormalizer(JavaBaseNormalizer):
    """Replaces all immediate values over some threshold with the immediate token
    
    Parameters
    ----------
    imm_threshold: `Optional[int]`
        all immediate values whose absolute value is greater than this threshold will be replaced with the immediate
        value token. If None or < 0, then all immediates will be replaced no matter the size
    include_negative: `bool`
        if True, then a negative sign will be added to the front of all replaced immediate tokens that are negative
    tokenizer: `Tokenizer`
        the tokenizer to use
    token_handlers: `Optional[Dict[str, Callable[[NormalizerState], Union[None, str]]]]`
        optional dictionary mapping string token types to functions to handle those tokens. These will override any
        token handlers that are used by default (IE: all of the `self.handle_*` functions). Functions should take one
        arg (the current normalizer state) as input and return either the next string token to add to the current line,
        or None to not add anything. This is useful for adding more methods to handle new token types that are not builtin.
    token_sep: `Optional[str]`
        the string to use to separate each token in returned instruction lines. Only used if tokenization_level is 
        'instruction'. If None, then a default value will be used (' ' for unnormalized using BaseNormalizer(), '_' 
        for everything else)
    tokenization_level: `Optional[Union[TokenizationLevel, str]]`
        the tokenization level to use for return values. Can be a string, or a ``TokenizationLevel`` type. Strings can be:

            - 'op': tokenized at the opcode/operand level. Will insert a 'INSTRUCTION_START' token at the beginning of
              each instruction line
            - 'inst'/'instruction': tokenized at the instruction level. All tokens in each instruction line are joined
              together using token_sep to construct the final token
            - 'auto': pick the default value for this normalization technique

    anonymize_tokens: `bool`
        if True, then tokens will be annonymized by taking their 4-byte shake_128 hash. Why does this exist? Bureaucracy.
    """
    def __init__(self, imm_threshold=None, include_negative=True, tokenizer=None, token_handlers=None, token_sep=None, 
                 tokenization_level=TokenizationLevel.AUTO, anonymize_tokens=False):
        super().__init__(tokenizer=tokenizer, token_handlers=token_handlers, token_sep=token_sep, 
                         tokenization_level=tokenization_level, anonymize_tokens=anonymize_tokens)

        self.imm_threshold = -1 if imm_threshold is None or imm_threshold < 0 else imm_threshold
        self.include_negative = include_negative

        self.handle_immediate = replace_immediate(include_negative=self.include_negative) if self.imm_threshold == -1 else \
            threshold_immediate(imm_threshold, include_negative)
    
    def __eq__(self, other):
        return super().__eq__(other) and self.imm_threshold == other.imm_threshold and self.include_negative == other.include_negative
    
    def __hash__(self):
        return super().__hash__() + hash_obj([self.imm_threshold, self.include_negative], return_int=True)



from setuptools import setup, find_packages

if __name__ == "__main__":
    setup(packages=find_packages(where='./src'), long_description=open('README.md').read(),
        long_description_content_type='text/markdown',)



import numpy as np
import pickle
import inspect
from enum import Enum
from ..normalization import normalize_cfg_data, Architectures
from .cfg import CFG, CFGBasicBlock, CFGFunction, EdgeType
from ..utils import get_smallest_np_dtype, scatter_nd_numpy, hash_obj, eq_obj, get_module, AtomicTokenDict
from ..utils.type_utils import *


if IN_PYTHON_TYPING_VERSION:
    BlockInfoBitDtype = np.int8  # Int type being used to store block info flags
    BlockInfoBitMaskFunctionType = Callable[[CFGBasicBlock], Literal[0, 1, True, False]]  # Dtype of functions that are used to compute bit flags


class MemCFG:
    """A CFG that is more memory/speed efficient.
    
    Keeps only the bare minimum information needed from a CFG. Stores edge connections in a CSR-like format.

    Parameters
    ----------
    cfg: `CFG`
        a CFG object. Can be a normalized or un-normalized. If un-normalized, then it will be normalized using the 
        `normalizer` parameter.
    normalizer: `Optional[Union[str, Normalizer]]`
        the normalizer to use to normalize the incoming CFG (or None if it is already normalized). If the incoming CFG 
        object has already been normalized, and `normalizer` is not None, then this will attempt to normalize the CFG 
        again with this normalizer
    keep_memory_addresses: `bool`
        if True, then memory addresses will also be kept. Otherwise they will be removed to save space
    inplace: `bool`
        if True and cfg needs to be normalized, it will be normalized inplace
    using_tokens: `Union[Dict[str, int], AtomicTokenDict]`
        if not None, then a dictionary mapping token strings to integer values. Any tokens in cfg but not in using_tokens
        will be added. Can also be an AtomicTokenDict for atomic updates to tokens
    force_renormalize: `bool`
        by default, this method will only normalize cfg's whose .normalizer != to the passed normalizer. However if 
        `force_renormalize=True`, then all cfg's will be renormalized even if they have been previously normalized with 
        the same normalizer.
    """

    normalizer: 'NormalizerType'
    """The normalizer used to normalize input before converting to ``MemCFG``
    
    Can be shared with a ``MemCFGDataset`` object if this ``MemCFG`` is a part of one
    """

    tokens: 'dict[str, int]'
    """Dictionary mapping token strings to integer values used in this ``MemCFG``
    
    Can be shared with a ``MemCFGDataset`` object if this ``MemCFG`` is a part of one.

    Can also be an AtomicTokenDict object for atomic token updates
    """

    function_name_to_idx: 'dict[str, int]'
    """Dictionary mapping string function names to their integer ids used in this ``MemCFG``"""

    asm_lines: 'np.ndarray'
    """Assembly line information
    
    A contiguous 1-d numpy array of shape (num_asm_lines,) of integer assembly line tokens. Dtype is the smallest 
    unsigned dtype needed to store the largest token value in this ``MemCFG``

    To get the assembly lines for some block index `block_idx`, you must get the assembly line indices from ``block_asm_idx``,
    and use those to slice the assembly lines:

    >>> block_idx = 7
    >>> memcfg.asm_lines[memcfg.block_asm_idx[block_idx]:memcfg.block_asm_idx[block_idx + 1]]

    Also see :func:`~bincfg.MemCFG.get_block_asm_lines`
    """

    asm_memory_addresses: 'Union[None, np.ndarray]'
    """Memory addresses for all of the assembly lines
    
    Only saved if `keep_memory_addresses=True` when constructing the ``MemCFG``. This will be a 1-d signed integer numpy
    array, where a value of -1 means the memory address for that corresponding line was not present in the basic block
    """

    block_asm_idx: 'np.ndarray'
    """Indices in ``asm_lines`` that correspond to the assembly lines for each basic block in this ``MemCFG``
    
    A 1-d numpy array of shape (num_blocks + 1,). Dtype is the smallest unsigned dtype needed to store the value 
    `num_asm_lines`. Assembly tokens for a block at index `i` would have a start index of `block_asm_idx[i]` and an end
    index of `block_asm_idx[i + 1]` in ``asm_lines``.
    """

    block_func_idx: 'np.ndarray'
    """Integer ids for the function that each basic block belongs to
    
    A 1-d numpy array of shape (num_blocks,) where each element is a function id for the block at that index. The id
    can be found in ``function_name_to_idx``. Dtype is the smallest unsigned dtype needed to store the value `num_functions`

    Also see :func:`~bincfg.MemCFG.get_block_function_idx` and :func:`~bincfg.MemCFG.get_block_function_name`
    """

    block_flags: 'np.ndarray'
    """Integer of bit flags for each basic block
    
    A 1-d numpy array of shape (num_blocks,) where each element is an integer of bit flags. See ``BlockInfoBitMask``
    for more info. Dtype is the smallest unsigned dtype with enough bits to store all flags in ``BlockInfoBitMask``

    Also see :func:`~bincfg.MemCFG.get_block_flags`
    """

    block_memory_addresses: 'Union[np.ndarray, None]'
    """Integer memory addresses of basic blocks.

    Only saved if `keep_memory_addresses=True` when constructing the ``MemCFG``. This will be a 1-d unsigned integer numpy
    array containing the memory addresses
    """

    block_asm_mem_addr_idx: 'Union[np.ndarray, None]'
    """Indices in ``block_memory_addresses`` that correspond to the assembly line memory addresses for basic blocks
    
    A 1-d numpy array of shape (num_blocks + 1,). Dtype is the smallest unsigned dtype needed to store the number of
    assembly line memory addresses. Memory addresses for a block at index `i` would have a start index of 
    `block_asm_mem_addr_idx[i]` and an end index of `block_asm_mem_addr_idx[i + 1]` in ``block_memory_addresses``.
    Only saved if `keep_memory_addresses=True` when constructing the ``MemCFG``.
    """

    metadata: 'dict'
    """Dictionary of metadata associated with this MemCFG"""

    function_metadata: 'list[Union[int, dict]]'
    """Metadata for functions
    
    A list of run length compressed metadata at the function level. We only compress metadata dictionaries that are empty.
    Elements are in the same order as the function indices in `block_func_idx`. Elements are either dictionaries (for the
    metadata of that current function), or integers indicating we should skip that many functions as they all have
    no metadata.
    """

    block_metadata: 'list[Union[int, dict]]'
    """Metadata for blocks
    
    A list of run length compressed metadata at the basic block level. We only compress metadata dictionaries that are empty.
    Elements are in the same order as the block indices in `block_asm_idx`. Elements are either dictionaries (for the
    metadata of that current block), or integers indicating we should skip that many blocks as they all have
    no metadata.
    """

    graph_c: 'np.ndarray'
    """Array containing all of the outgoing edges for each block in order
    
    1-D numpy array of shape (num_edges,). Dtype will be the smallest unsigned dtype required to store the value
    `num_blocks + 1`. Each element is a block index to which that edge connects. Edges will be in the order they appear in 
    each block's ``edges_out`` attribute, for each block in order of their block_idx. 

    Also see :func:`~bincfg.MemCFG.get_edges_out`

    NOTE: this also contains information on which types of edges they are inherently. If the block is NOT a function call
    (stored as bit flag in the block_info array), then all edges for that block are normal edges. If it IS a function
    call, then there are 3 cases:

        1. it has one outgoing edge: that edge is always a function call
        2. it has two outgoing edges, one function call, one normal: the first edge is the function call edge, the second
           is a normal edge
        3. it has >2 outgoing edges, or 2 function call edges: the edges will be listed first by function call edges, 
           then by normal edges, with a separator inbetween. The separator will have the max unsigned int value for 
           graph_c's dtype. This is why we use the dtype that can store `num_blocks + 1`, since we need this extra value
           just in case. Whatever exactly it means for a basic block to have >2 outgoing edges while being a function 
           call is left up to the user. Possibly due to call operators with non-explicit operands (eg: register memory 
           locations)?
    """

    graph_r: 'np.ndarray'
    """Array containing information on the number of outgoing edges for each block
    
    1-D numpy array of shape (num_edges + 1,). Dtype will be the smallest unsigned dtype required to store the value
    `num_edges`. This array is a cumulative sum of the number of edges for each basic block. One could get all of the 
    outgoing edges for a block using:

    >>> start_idx = memcfg.graph_r[block_idx]
    >>> end_idx = memcfg.graph_r[block_idx + 1]
    >>> edges = memcfg.graph_c[start_idx:end_idx]

    Also see :func:`~bincfg.MemCFG.get_edges_out`
    """

    class BlockInfoBitMask(Enum):
        """An Enum for block info bit masks
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                        
        Each value is a tuple of the bit mask for that boolean, and a function to call with the block that returns a
        boolean True if that bit should be set, False otherwise. If True, then that bit will be '1' in that block's 
        block_flags int.
        """

        IS_FUNCTION_CALL: 'Tuple[int, BlockInfoBitMaskFunctionType]' = (1 << 0, lambda block: block.is_function_call)
        """Bit set if this block is a function call. See :py:func:`~bincfg.cfg_basic_block.is_function_call`"""
        IS_FUNCTION_ENTRY: 'Tuple[int, BlockInfoBitMaskFunctionType]' = (1 << 1, lambda block: block.is_function_entry)
        """Bit set if this block is a function entry. See :py:func:`~bincfg.cfg_basic_block.is_function_entry`"""
        IS_IN_EXTERN_FUNCTION: 'Tuple[int, BlockInfoBitMaskFunctionType]' = (1 << 2, lambda block: block.parent_function.is_extern_function)
        """Bit set if this block is within an external function. See :py:func:`~bincfg.cfg_function.is_extern_function`"""
        IS_FUNCTION_JUMP: 'Tuple[int, BlockInfoBitMaskFunctionType]' = (1 << 3, lambda block: block.is_function_jump)
        """Bit set if this block is a function jump. IE: this block has a jump instruction that resolves to a basic block
        in a separate function. See :py:func:`~bincfg.cfg_basic_block.is_function_jump`"""
        IS_MULTI_FUNCTION_CALL: 'Tuple[int, BlockInfoBitMaskFunctionType]' = (1 << 4, lambda block: 0)
        """Bit set if this block is a multi-function call. IE: this block has either two or more function call edges out,
        or one function call and two or more normal edges out. See :py:func:`~bincfg.cfg_basic_block.is_multi_function_call`
        
        Currently not setting the block here in _block_flags_int(), but instead in MemCFG initialization in order to save
        time (we don't have to compute get_sorted_edges() multiple times)
        """

    @staticmethod
    def _block_flags_int(block: 'CFGBasicBlock') -> 'int':
        """Gets all of the block information and stores it as a integer of bit-set flags

        Args:
            block (CFGBasicBlock): the CFGBasicBlock to get the info int from

        Returns:
            int: the integer of bit-set flags
        """
        block_flags = 0
        for bm in MemCFG.BlockInfoBitMask:
            block_flags |= bm.value[0] * bm.value[1](block)  # Times will convert boolean to 1 or 0
        return block_flags

    def __init__(self, cfg: 'CFG', normalizer: 'Optional[Union[str, NormalizerType]]' = None, keep_memory_addresses: 'bool' = False,
                 inplace: 'bool' = False, using_tokens: 'Optional[Union[dict, AtomicTokenDict]]' = None, force_renormalize: 'bool' = False):
        # Make sure input is a cfg
        if not isinstance(cfg, CFG):
            raise TypeError("Can only build a MemCFG out of a CFG object, not '%s'" % type(cfg).__name__)

        # Normalize the CFG if needed
        if normalizer is not None or cfg.normalizer is None:
            # Make sure there is some normalizer to use
            if normalizer is None:
                raise ValueError("Must pass a normalizer if `cfg` is unnormalized!")
            cfg = normalize_cfg_data(cfg, normalizer=normalizer, inplace=inplace, force_renormalize=force_renormalize)
        
        # Keep cfg's normalization if possible
        self.normalizer = cfg.normalizer

        # Figure out what the tokens should be, updating the token dict if we find new ones
        self.tokens = {} if using_tokens is None else using_tokens

        if isinstance(self.tokens, AtomicTokenDict):
            self.tokens.addtokens(*{l for block in cfg.blocks for l in block.asm_lines})
        else:
            for block in cfg.blocks:
                for l in block.asm_lines:
                    self.tokens.setdefault(l, len(self.tokens))
                    
        max_token = len(self.tokens)
        max_asm_addr = default_max((a for b in cfg.blocks for a in b.asm_memory_addresses), 0)
        max_block_addr = default_max((b.address for b in cfg.blocks), 0)

        # Make mappings from function names to indices. Make sure there aren't duplicates (most likely only going to
        #   occur in functions with no name)
        self.function_name_to_idx = {}
        function_addr_to_idx = {}
        self.function_metadata = []
        for i, f in enumerate(cfg.functions):
            if f.name in self.function_name_to_idx:
                func_name = f.name
                fn_idx = 0
                while func_name in self.function_name_to_idx:
                    func_name = f.name + '_%d' % fn_idx
                    fn_idx += 1
            else:
                func_name = f.name
            self.function_name_to_idx[func_name] = i
            function_addr_to_idx[f.address] = i
            self.function_metadata.append(f.metadata)

        # Make the data arrays
        self.asm_lines = np.empty([cfg.num_asm_lines], dtype=get_smallest_np_dtype(max_token))
        self.block_asm_idx = np.empty([cfg.num_blocks + 1], dtype=get_smallest_np_dtype(cfg.num_asm_lines))
        self.block_func_idx = np.empty([cfg.num_blocks], dtype=get_smallest_np_dtype(len(cfg.functions_dict)))
        self.block_flags = np.empty([cfg.num_blocks], dtype=get_smallest_np_dtype(len(MemCFG.BlockInfoBitMask)))
        graph_c = []
        self.graph_r = np.empty([cfg.num_blocks + 1])
        self.block_labels = {}  # Mapping of block indices to block labels
        self.block_metadata = []

        # Set the initial graph_r start and final block_asm_idx
        self.graph_r[0] = 0
        self.block_asm_idx[-1] = len(self.asm_lines)

        if keep_memory_addresses:
            self.block_memory_addresses = np.empty([cfg.num_blocks], dtype=get_smallest_np_dtype(max_block_addr, signed=True))
            self.asm_memory_addresses = np.empty(sum(len(b.asm_memory_addresses) for b in cfg.blocks), dtype=get_smallest_np_dtype(max_asm_addr, signed=True))
            self.block_asm_mem_addr_idx = np.empty([cfg.num_blocks + 1], dtype=get_smallest_np_dtype(cfg.num_asm_lines))
            self.block_asm_mem_addr_idx[-1] = len(self.asm_memory_addresses)
        else:
            self.asm_memory_addresses, self.block_asm_mem_addr_idx, self.block_memory_addresses = None, None, None

        # Copy the metadata from the cfg
        self.metadata = cfg.metadata.copy()

        # Create temporary mapping of CFGBasicBlock.address to integer index of block (sorted by address for determinism)
        block_addr_to_idx = {block.address: i for i, block in enumerate(cfg.blocks)}

        # Convert the data at each block to better memory one
        asm_line_idx, asm_mem_idx = 0, 0
        for block_idx, block in enumerate(cfg.blocks):

            # Get the assembly lines and memory addresses, and store the length of the assembly lines
            asm_line_end = asm_line_idx + block.num_asm_lines
            self.asm_lines[asm_line_idx: asm_line_end] = [self.tokens[l] for l in block.asm_lines]
            self.block_asm_idx[block_idx] = asm_line_idx
            asm_line_idx = asm_line_end

            # Add in the memory addresses if using
            if self.asm_memory_addresses is not None:
                self.block_memory_addresses[block_idx] = block.address
                asm_mem_end = asm_mem_idx + len(block.asm_memory_addresses)
                self.asm_memory_addresses[asm_mem_idx: asm_mem_end] = block.asm_memory_addresses
                self.block_asm_mem_addr_idx[block_idx] = asm_mem_idx
                asm_mem_idx = asm_mem_end

            # Get the block's function name
            self.block_func_idx[block_idx] = function_addr_to_idx[block.parent_function.address]

            # Get all of the edges associated with this block in order: normal edges, function call edges
            edge_lists = block.get_sorted_edges(direction='out', as_sets=False)

            # Add in the function call edges
            graph_c += [block_addr_to_idx[edge.to_block.address] for edge in edge_lists[1]]

            # Check for either >1 function call edges, or >2 edges while being a function call.
            # We add a -1, which will be rolled over to the unsigned int max value for graph_c array
            added_minus_1 = 0
            if len(edge_lists[1]) >= 2 or (len(edge_lists[1]) == 1 and len(edge_lists[0]) > 1):
                graph_c.append(-1)
                added_minus_1 = 1
            
            # Add in the normal edges
            graph_c += [block_addr_to_idx[edge.to_block.address] for edge in edge_lists[0]]

            # Update the new graph_r, taking into account whether or not we inserted an extra -1 to split function call/normal edges
            self.graph_r[block_idx + 1] = sum(len(l) for l in edge_lists) + added_minus_1
            
            # Get the block information flags for this block.
            # Set the flag for is_multi_function_call here so we don't have to call get_sorted_edges more than once
            self.block_flags[block_idx] = MemCFG._block_flags_int(block) | \
                (MemCFG.BlockInfoBitMask.IS_MULTI_FUNCTION_CALL.value[0] * added_minus_1)
            
            # Get the block metadata and memory address
            self.block_metadata.append(block.metadata)
        
        # Conver graph_c and graph_r to numpy arrays
        graphc_dtype = get_smallest_np_dtype(len(cfg.blocks_dict) + 1, signed=False)
        self.max_graph_c_int = np.iinfo(graphc_dtype).max
        self.graph_c = np.array([(v if v != -1 else self.max_graph_c_int) for v in graph_c], dtype=graphc_dtype)
        self.graph_r = np.cumsum(self.graph_r, dtype=get_smallest_np_dtype(len(self.graph_c)))
        
        # Run length encode the metadata
        self.function_metadata = _rl_encode_metadata(self.function_metadata)
        self.block_metadata = _rl_encode_metadata(self.block_metadata)
    
    def get_block_info(self, block_idx):
        """Returns all the info associated with the given block index as a dictionary

        Args:
            block_idx (int): integer block index

        Returns:
            dict: the block info dictionary with keys/values:

                * 'asm_lines' (np.ndarray): 1-d numpy array of unsigned integer assembly line tokens in this block
                * 'asm_memory_addresses' (np.ndarray): 1-d numpy array of signed integer memory addresses
                  for the assembly lines in this block. Values will be -1 if the memory addresses do not exist
                * 'edges_out' (np.ndarray): 1-d numpy array of unsigned integer block indices for all
                  of the edges out from this block
                * 'edge_types' (np.ndarray): 1-d numpy array of uint8 values for the edge types associated
                  with all of the edges out. These are the values of objects in the EdgeType enum. Currently: EdgeType.NORMAL == 1,
                  EdgeType.FUNCTION_CALL == 2
                * 'function_index' (int): the integer function index of the function this block resides in
                * 'is_function_call' (bool): true if this block is a function call block (has at least one outgoing function call edge)
                * 'is_function_entry' (bool): true if this block is a function entry block (has the same memory address
                  as its parent function)
                * 'is_extern_function' (bool): true if this block is within an external function (parent_function.is_extern_function is True)
                * 'is_function_jump' (bool): true if this block is a function jump block (has a 'normal' edge to a block
                  that is within another function)
                * 'is_multi_function_call' (bool): true if this block is a multi-function call block (has 2 or more outgoing
                  function call edges. IE: a call table)
                * 'metadata' (dict): dictionary of metadata associated with this block
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')

        ret = (self.get_block_asm_lines(block_idx),) + (self.get_block_asm_memory_addresses(block_idx),) + self.get_block_edges_out(block_idx, ret_edge_types=True) \
            + (self.get_block_function_idx(block_idx),) + self.get_block_flags(block_idx) + (self.get_block_metadata(block_idx),)
        
        ret = {k: v for k, v in zip(['asm_lines', 'asm_memory_addresses', 'edges_out', 'edge_types', 'function_index', 'is_function_call', 
                'is_function_entry', 'is_extern_function', 'is_function_jump', 'is_multi_function_call', 'metadata'], ret)}
        
        return ret

    def get_block_asm_lines(self, block_idx: 'int') -> 'np.ndarray':
        """Get the asm lines associated with this block index

        Args:
            block_idx (int): integer block index

        Returns:
            np.ndarray: a 1-d numpy array of unsigned integer assembly tokens
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return self.asm_lines[self.block_asm_idx[block_idx]:self.block_asm_idx[block_idx+1]]

    def get_block_asm_memory_addresses(self, block_idx: 'int') -> 'np.ndarray':
        """Get the asm memory addresses associated with this block index

        Values are -1 if the memory address did not exist in that block

        Args:
            block_idx (int): integer block index

        Returns:
            np.ndarray: a 1-d numpy array of signed integer assembly tokens
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return self.asm_memory_addresses[self.block_asm_mem_addr_idx[block_idx]:self.block_asm_mem_addr_idx[block_idx+1]]
    
    def get_block_edges_out(self, block_idx: 'int', ret_edge_types: 'bool' = False) -> \
        'Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]':
        """Get numpy array of block indices for all edges out associated with the given block index

        Args:
            block_idx (int): integer block index
            ret_edge_types (bool): if True, will also return a numpy array (1-d, dtype np.uint8) containing the edge
                type values for each edge with values:

                    - 1: normal edge
                    - 2: function call edge

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray: 
                either a 1-d numpy array of unsigned integer block indices for all edges out associated with the given 
                block index, or if `ret_edge_types=True`, then a tuple of (block_edge_inds, edge_types) where the `edge_types`
                is a 1-d numpy array of uint8 edge types with the same shape as block_edge_inds that designates the types
                of the edges. Edge types will be the values of those in the `EdgeType` enum.
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')

        # Get all of the edges
        ret = self.graph_c[self.graph_r[block_idx]:self.graph_r[block_idx+1]]

        # Check if we are returning the edge types as well
        if ret_edge_types:

            # Check if this block is a function call
            if self.is_block_function_call(block_idx):

                # Check to see if this block is a multi-function call, in which case we have to split up the array
                if self.is_block_multi_function_call(block_idx):
                    
                    # Get the split index (index in ret that is equal to unsigned dtype max). It should have split_idx
                    #   function call edges, and len(ret) - split_idx - 1 normal edges (to account for the fact that
                    #   the split_idx itself is also stored in the array)
                    split_idx = np.argwhere(ret == self.max_graph_c_int)[0][0]
                    return ret[ret != self.max_graph_c_int], \
                        np.array([EdgeType.FUNCTION_CALL.value] * split_idx + [EdgeType.NORMAL.value] * (len(ret) - split_idx - 1))
                
                # Otherwise it is not a multi-function call. We can simply return ret, and edge types are either 
                #   [function_call] or [function_call, normal]
                else:
                    return ret, np.array(([EdgeType.FUNCTION_CALL.value] if len(ret) == 1 else \
                        [EdgeType.FUNCTION_CALL.value, EdgeType.NORMAL.value]), dtype=np.uint8)
            
            # It's not a function call, so we can just return ret and all edge types must be normal edges
            else:
                return ret, np.full(ret.shape, EdgeType.NORMAL.value, dtype=np.uint8)

        # Otherwise we are not returning the edge types, just return all values in ret that are not the splitting value
        # We only need to remove the splitting value if there are more than two edges, AND it is a function call. Otherwise
        #   we are doing a useless numpy comparison and memory-grabbing every call which takes a non-neglible amount of time
        return ret[ret != self.max_graph_c_int]

    def get_block_function_idx(self, block_idx: 'int') -> 'int':
        """Get the function index for the given block index

        Args:
            block_idx (int): integer block index

        Returns:
            int: the integer function index for the given block index
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return self.block_func_idx[block_idx]
    
    def get_block_function_name(self, block_idx: 'int') -> 'str':
        """Get the function name for the given block index
        
        Functions without names will start with '__unnamed_func__'

        Args:
            block_idx (int): integer block index

        Returns:
            str: the function name for the given block index
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')

        func_idx = self.get_block_function_idx(block_idx)

        # Make an inverse mapping now that we know we are calling this function
        if not hasattr(self, 'function_idx_to_name'):
            self.function_idx_to_name = {v: k for k, v in self.function_name_to_idx.items()}

        return self.function_idx_to_name[func_idx]
    
    def get_block_memory_address(self, block_idx: 'int') -> 'int':
        """Returns the memory address for the given block, if present, -1 if not present
        
        Args:
            block_idx (int): integer block index
        
        Returns:
            int: the memory address
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return -1 if self.block_memory_addresses is None else self.block_memory_addresses[block_idx]
    
    def get_block_flags(self, block_idx: 'int') -> 'Tuple[bool, bool, bool, bool, bool, bool]':
        """Get all block flags for the given block index

        Args:
            block_idx (int): integer block index

        Returns:
            Tuple[bool, bool, bool, bool, bool, bool]: (is_block_function_call, is_block_function_entry, 
                is_block_extern_function, is_block_function_jump, is_block_multi_function_call)
        """
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')

        return self.is_block_function_call(block_idx), self.is_block_function_entry(block_idx), \
            self.is_block_extern_function(block_idx), self.is_block_function_jump(block_idx), \
            self.is_block_multi_function_call(block_idx)
    
    def get_function_block_inds(self, func_idx: 'int') -> 'list[int]':
        """Returns all of the block indices that are within the given function
        
        Args:
            func_idx (int): the integer function index
        
        Returns:
            list[int]: list of integer block indices that are within the given function
        """
        assert_valid_idx(func_idx, self.num_functions, 'functions')

        if func_idx not in self.function_name_to_idx.values():
            raise ValueError("Unknown function index: %d" % func_idx)
        return [i for i in range(self.num_blocks) if self.block_func_idx[i] == func_idx]

    def get_function_metadata(self, func_idx: 'Union[int, None]') -> 'Union[dict, list[dict]]':
        """Returns the metadata associated with that function index
        
        Args:
            func_idx (Union[int, None]): the integer function index of the metadata to get, or None to get the full
                list of metadata
        
        Returns:
            Union[dict, list[dict]]: dictionary of metadata associated with the given function index
        """
        if isinstance(func_idx, INTEGER_TYPES):
            assert_valid_idx(func_idx, self.num_functions, 'functions')
        return _decode_rl_metadata(self.function_metadata, func_idx)

    def get_block_metadata(self, block_idx: 'Union[int, None]') -> 'Union[dict, list[dict]]':
        """Returns the metadata associated with that function index
        
        Args:
            block_idx (Union[int, None]): the integer block index of the metadata to get, or None to get the full
                list of metadata
        
        Returns:
            Union[dict, list[dict]]: dictionary of metadata associated with the given block index
        """
        if isinstance(block_idx, INTEGER_TYPES):
            assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return _decode_rl_metadata(self.block_metadata, block_idx)
    
    def is_block_function_call(self, block_idx: 'int') -> 'bool':
        """True if this block is a function call, False otherwise"""
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return (self.block_flags[block_idx] & MemCFG.BlockInfoBitMask.IS_FUNCTION_CALL.value[0]) > 0
    
    def is_block_function_entry(self, block_idx: 'int') -> 'bool':
        """True if this block is a function entry, False otherwise"""
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return (self.block_flags[block_idx] & MemCFG.BlockInfoBitMask.IS_FUNCTION_ENTRY.value[0]) > 0
    
    def is_block_extern_function(self, block_idx: 'int') -> 'bool':
        """True if this block is in an external function, False otherwise"""
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return (self.block_flags[block_idx] & MemCFG.BlockInfoBitMask.IS_IN_EXTERN_FUNCTION.value[0]) > 0
    
    def is_block_function_jump(self, block_idx: 'int') -> 'bool':
        """True if this block is a function jump, False otherwise"""
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return (self.block_flags[block_idx] & MemCFG.BlockInfoBitMask.IS_FUNCTION_JUMP.value[0]) > 0
    
    def is_block_multi_function_call(self, block_idx: 'int') -> 'bool':
        """True if this block is a multi-function call, False otherwise"""
        assert_valid_idx(block_idx, self.num_blocks, 'blocks')
        return (self.block_flags[block_idx] & MemCFG.BlockInfoBitMask.IS_MULTI_FUNCTION_CALL.value[0]) > 0
    
    @property
    def num_blocks(self) -> 'int':
        """The number of blocks in this ``MemCFG``"""
        return len(self.block_func_idx)
    
    @property
    def num_edges(self) -> 'int':
        """The number of edges in this ``MemCFG``"""
        return len(self.graph_c[self.graph_c != self.max_graph_c_int])
    
    @property
    def num_asm_lines(self) -> 'int':
        """The number of assembly lines in this ``MemCFG``"""
        return len(self.asm_lines)
    
    @property
    def num_functions(self) -> 'int':
        """The number of functions in this ``MemCFG``"""
        return len(self.function_name_to_idx)
    
    @property
    def inv_tokens(self) -> 'dict[int, str]':
        """Returns the inverse of `self.tokens`: dictionary mapping token integers to their original strings"""
        return {v:k for k, v in self.tokens.items()}
    
    def update_metadata(self, other: 'dict') -> 'Self':
        """Updates this MemCFG's metadata dictionary with the given dictionary, and returns self"""
        self.metadata.update(other)
        return self
    
    def set_tokens(self, tokens: 'Union[dict, AtomicTokenDict]') -> 'Self':
        """Sets this MemCFG's tokens to the given tokens, and returns self"""
        self.tokens = tokens
        return self
    
    def normalize(self, normalizer: 'Optional[Union[str, NormalizerType]]' = None, using_tokens: 'Union[dict, AtomicTokenDict]' = None, 
                  inplace: 'bool' = True, force_renormalize: 'bool' = False) -> 'MemCFG':
        """Normalizes this memcfg in-place.

        Args:
            normalizer (Optional[Union[str, NormalizerType]]): the normalizer to use. Can be a ``Normalizer`` object, or a 
                string, or None to use the default BaseNormalizer(). Defaults to None.
            using_tokens (Union[dict, AtomicTokenDict]): tokens to use when normalizing
            inplace (bool): whether or not to normalize inplace. Defaults to True.
            force_renormalize (bool): by default, this method will only normalize this cfg if the passed 
                `normalizer` is != `self.normalizer`. However if `force_renormalize=True`, then this will be renormalized
                even if it has been previously normalized with the same normalizer. Defaults to False.

        Returns:
            MemCFG: this ``MemCFG`` normalized
        """
        return normalize_cfg_data(self, normalizer=normalizer, using_tokens=using_tokens, inplace=inplace, 
                                  force_renormalize=force_renormalize)
    
    @property
    def architecture(self) -> 'Architectures':
        """Returns the architecture being used. Currently a WIP
        
        Checks for an 'arch' or 'architecture' key in the metadata and returns it if it is known. Can currently return:
        'java', 'x86'
        """
        for k in ['arch', 'architecture']:
            if k in self.metadata:
                arch = self.metadata[k]
                break
        else:
            raise ValueError("Could not find 'arch' or 'architecture' key in metadata")
        
        if arch in ['x86']:
            return Architectures.X86
        elif arch in ['java', 'java-bytecode']:
            return Architectures.JAVA
        else:
            raise ValueError("Unknown architecture value: %s" % repr(arch))
    
    def get_edge_values(self) -> 'np.ndarray':
        """Returns the edge type values
        
        Returns a 1-d numpy array of length self.num_edges and dtype np.int32 containing an integer type for each
        edge depending on if it is a normal or function call edge. Edges are directed and have values from `EdgeType` enum.
        Values:

            - 1: 'normal' edges
            - 2: 'function call' edges
        
        NOTE: this returns as type np.int32 since pytorch can be finicky about what dtypes it wants

        Returns:
            np.ndarray: a 1-d numpy array of length self.num_edges and dtype np.int32 containing integer edge types
        """
        return np.array([v for b in range(self.num_blocks) for v in self.get_block_edges_out(b, ret_edge_types=True)[1]], dtype=np.int32)
    
    def get_coo_indices(self) -> 'np.ndarray':
        """Returns the COO indices for this MemCFG

        Returns a 2-d numpy array of shape (num_edges, 2) of dtype np.int32. Each row is an edge, column 0 is the 'row' 
        indexer, and column 1 is the 'column' indexer. EG:
        
        .. highlight:: python
        .. code-block:: python
        
            original = np.array([
                [0, 1],
                [1, 1]
            ])

            coo_indices = np.array([
                [0, 1],
                [1, 0],
                [1, 1]
            ])
        
        NOTE: this returns as type np.int32 since pytorch can be finicky about what dtypes it wants
        NOTE: pytorch sparse_coo_tensor's indicies are the transpose of the array this method returns

        Returns:
            np.ndarray: a 2-d numpy array of shape (num_edges, 2) of dtype np.int32 containing COO indices
        """
        inds = np.empty([self.num_edges, 2], dtype=np.int32)
        start = 0
        for bi in range(self.num_blocks):
            edges_out = self.get_block_edges_out(bi, ret_edge_types=False)
            end = start + len(edges_out)
            inds[start:end, 0] = bi
            inds[start:end, 1] = edges_out
            start = end
        return inds
    
    def to_adjacency_matrix(self, type: 'Literal["np", "numpy", "torch"]' = 'np', sparse: 'bool' = False) ->\
        'Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]':
        """Returns an adjacency matrix representation of this memcfg's graph connections

        Edges are directed and have values from `EdgeType` enum. Values:

            - 1: 'normal' edges
            - 2: 'function call' edges

        Args:
            type (Literal["np", "numpy", "torch"]): the type of matrix to return. Defaults to 'np'. Can be:

                - 'np'/'numpy' for a numpy ndarray (dtype: np.int32)
                - 'torch'/'pytorch' for a pytorch tensor (type: LongTensor)
            
            sparse (bool): whether or not the return value should be a sparse matrix. Defaults to False. Has 
                different behaviors based on type:

                - numpy array: returns a 2-tuple of sparse COO representation (indices, values).
                    NOTE: the indices are the transpose of those from `get_coo_indices()`
                    NOTE: if you want sparse CSR format, you already have it with self.graph_c and self.graph_r
                - pytorch tensor: returns a pytorch sparse COO tensor. 
                    NOTE: not using sparse CSR format for now since it seems to have less documentation/supportedness. 

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray: 
                an adjacency matrix representation of this ``MemCFG``
        """
        type = type.lower()

        # Return adj mat as intended type
        if type in ['np', 'numpy']:
            if sparse:
                return (self.get_coo_indices().T, self.get_edge_values())
            else:
                ret = np.zeros((self.num_blocks, self.num_blocks))
                return scatter_nd_numpy(ret, self.get_coo_indices(), self.get_edge_values())

        elif type in ['torch', 'pytorch']:
            torch = get_module('torch', err_message="Cannot find module `torch` required to return pytorch tensors!")
            sparse_coo = torch.sparse_coo_tensor(indices=self.get_coo_indices().T, values=self.get_edge_values(), size=(self.num_blocks, self.num_blocks)).coalesce()
            return sparse_coo if sparse else sparse_coo.to_dense()

        else:
            raise ValueError("Unknown adjacency matrix type: '%s'" % type)
    
    def to_cfg(self) -> 'CFG':
        """Converts this MemCFG back into a CFG
        
        NOTE: if `keep_memory_addresses=False` when constucting this MemCFG, then memory addresses will not be present
        and basic blocks will be given a memory address that is just their index in the block list
        """
        cfg = CFG(metadata=self.metadata, normalizer=self.normalizer)

        # Build the blocks
        edge_types = [None] * (max(et.value for et in EdgeType) + 1)
        for et in EdgeType:
            edge_types[et.value] = et.name.lower()

        block_metadata = self.get_block_metadata(None)
        block_addrs = self.block_memory_addresses if self.block_memory_addresses is not None else np.arange(self.num_blocks)

        blocks = [CFGBasicBlock(
            address = block_addrs[block_idx],
            edges_out = [(None, block_addrs[addr], edge_types[et]) for addr, et in zip(*self.get_block_edges_out(block_idx, ret_edge_types=True))], 
            asm_lines = [self.inv_tokens[t] for t in self.asm_lines[self.block_asm_idx[block_idx]:self.block_asm_idx[block_idx+1]]],
            asm_memory_addresses = self.asm_memory_addresses[self.block_asm_mem_addr_idx[block_idx]:self.block_asm_mem_addr_idx[block_idx+1]]\
                if self.asm_memory_addresses is not None else None,
            metadata = block_metadata[block_idx],
        ) for block_idx in range(self.num_blocks)]

        # Build the functions
        func_metadata = self.get_function_metadata(None)
        funcs = [CFGFunction(metadata=func_metadata[func_idx]) for func_idx in range(self.num_functions)]
        for block_idx in range(self.num_blocks):
            func_idx = self.get_block_function_idx(block_idx)

            funcs[func_idx].blocks.append(blocks[block_idx])
            funcs[func_idx].name = self.get_block_function_name(block_idx)
            if self.is_block_extern_function(block_idx):
                funcs[func_idx]._is_extern_function = True
            if self.is_block_function_entry(block_idx):
                funcs[func_idx].address = blocks[block_idx].address
        
        # Add the functions to the cfg and return
        cfg.add_function(*funcs)
        return cfg
    
    def save(self, path: 'str') -> 'None':
        """Saves this MemCFG to the given path"""
        with open(path, 'wb') as f:
            pickle.dump(self, f)
    
    def dumps(self) -> 'str':
        """Returns this object pickled with pickle.dumps()"""
        return pickle.dumps(self)
    
    def drop_tokens(self) -> 'Self':
        """Sets the tokens in this normalizer to None. Make sure you only do this if tokens are saved elsewhere! Returns self"""
        self.tokens = None
        return self
    
    @classmethod
    def load(cls, path: 'str') -> 'MemCFG':
        """Loads a MemCFG from the given path"""
        with open(path, 'rb') as f:
            return pickle.load(f)
        
    def __str__(self) -> 'str':
        return "MemCFG with normalizer: %s and %d functions, %d blocks, %d assembly lines, and %d edges. Metadata:\n%s" % \
            (repr(str(self.normalizer)), self.num_functions, self.num_blocks, self.num_asm_lines, self.num_edges, self.metadata)
    
    def __repr__(self) -> 'str':
        return self.__str__()

    def __getstate__(self) -> 'dict':
        """State for pickling"""
        ret = self.__dict__.copy()
        if 'function_idx_to_name' in ret:
            del ret['function_idx_to_name']
        return ret
    
    def __setstate__(self, state: 'dict') -> 'None':
        """Set state for pickling"""
        for k, v in state.items():
            setattr(self, k, v)

        if not hasattr(self, 'max_graph_c_int'):
            self.max_graph_c_int = np.iinfo(self.graph_c.dtype).max
    
    def __eq__(self, other) -> 'bool':
        return isinstance(other, MemCFG) and all(eq_obj(self, other, selector=s) for s in [
            'normalizer', 'tokens', 'function_name_to_idx', 'asm_lines', 'asm_memory_addresses', 'block_asm_idx', 'block_func_idx', 
            'block_flags',  'metadata', 'function_metadata', 'block_metadata', 'graph_c', 'graph_r', 'block_labels'
        ])
    
    def __hash__(self) -> 'hash':
        return hash_obj([
            self.normalizer, self.tokens, self.function_name_to_idx, self.asm_lines, self.asm_memory_addresses, 
            self.block_asm_idx, self.block_func_idx, self.block_flags, self.metadata, self.function_metadata, 
            self.block_metadata, self.graph_c, self.graph_r, self.block_labels,
        ], return_int=True)
    

def _rl_encode_metadata(meta_list: 'list[dict]') -> 'list[Union[int, dict]]':
    """Run length encodes metadata. Only encodes empty dictionaries for now"""
    ret = []
    for d in meta_list:
        if d is None or len(d) == 0:
            # If we need to insert a new integer for rl encoding
            if len(ret) == 0 or isinstance(ret[-1], dict):
                ret.append(1)
            else:
                ret[-1] += 1
        else:
            ret.append(d)
    return ret


def _decode_rl_metadata(meta_list: 'list[Union[int, dict]]', idx: 'Optional[int]' = None):
    """Decodes run length encoded metadata lists
    
    Assumes 0 < `idx` < number of functions

    Args:
        meta_list (list[Union[int, dict]]): the metadata compressed list
        idx (Optional[int]): if None, then this will return the full uncompressed list. Otherwise this will return the
            element at this index
    """
    # If we are just returning the full list of metadata
    if idx is None:
        ret = []
        for e in meta_list:
            if isinstance(e, int):
                ret += [{} for _ in range(e)]
            else:
                ret.append(e)
        return ret

    assert 0 <= idx < len(meta_list) + sum((e - 1) for e in meta_list if isinstance(e, int))

    # Otherwise, we can stop early if possible
    curr_idx = 0
    ml_idx = 0
    while True:
        if curr_idx == idx:
            return {} if isinstance(meta_list[ml_idx], int) else meta_list[ml_idx]
        elif isinstance(meta_list[ml_idx], int):
            # If our index is in this range, return empty dict
            if curr_idx <= idx < curr_idx + meta_list[ml_idx]:
                return {}
            curr_idx += meta_list[ml_idx]
        else:
            curr_idx += 1
        ml_idx += 1

    # We should not be able to reach here
    assert False


def assert_valid_idx(idx: 'int', max_val: 'int', objects_str: 'str') -> 'None':
    """Asserts that the idx passed is >= 0 and < max_val. `objects_str` is the type of object for error message (IE: 'blocks', 'functions', etc.)"""
    func_name = repr(inspect.getframeinfo(inspect.currentframe().f_back)[2])
    if idx < 0:
        raise ValueError("Index passed to function %s must be non-negative, got: %d" % (func_name, idx))
    elif idx >= max_val:
        raise ValueError("Index passed to function %s must be < the maximum number of %s (%d), got: %d" % (func_name, objects_str, max_val, idx))


def default_max(iterable, default=None):
    """Returns the max value in `iterable`, or `default` value if len(iterable) == 0"""
    try:
        return max(iterable)
    except ValueError:
        return default
    


import re
import bincfg
from .x86_tokenizer import Tokens, X86_REGISTER_SIZES, X86_MEMORY_SIZES
from ..norm_utils import imm_to_int, IMMEDIATE_VALUE_STR, GENERAL_REGISTER_STR, MEM_SIZE_TOKEN_STR, JUMP_DESTINATION_STR, \
    MEMORY_EXPRESSION_STR, FUNCTION_CALL_STR, MULTI_FUNCTION_CALL_STR, RECURSIVE_FUNCTION_CALL_STR, EXTERNAL_FUNCTION_CALL_STR, \
    INTERNAL_FUNCTION_CALL_STR, DISPLACEMENT_IMMEDIATE_STR, STRING_LITERAL_STR, RE_SPACING


# Regex's to check if registers are general or not, and to remove number information.
RE_GENERAL_REGISTER_MATCH = re.compile(r'r[0-9]+[dwb]?|[re]?[abcd]x|[abcd][lh]|[re]?[sd]il?', flags=re.IGNORECASE)
RE_REMOVE_REGISTER_NUMBER = re.compile(r'\(?[0-9]+\)?')

# Handling memory size information
MEM_SIZE_RE = re.compile(r'(?:v([0-9]+))?([a-z]+)(?:{space}ptr)?'.format(space=RE_SPACING), flags=re.IGNORECASE)


def x86_clean_nop(state):
    """Cleans any line with the opcode 'nop' to only contain the opcode

    Args:
        idx (int): the index in ``line`` of the 'nop' opcode
        line (List[TokenTuple]): a list of (token_type, token) tuples. the current assembly line
        args: unused
        kwargs: unused

    Returns:
        int: integer index in line of last handled token
    """
    old_line = ' '.join([l[2] for l in state.line])
    state.line.clear()
    state.line.append((Tokens.OPCODE, 'nop', old_line))
    return 1


def x86_replace_general_register(self, state):
    """Replaces general registers with a default string and their size, keeping special registers the same (while removing their numbers)

    Args:
        token (str): the current string token

    Returns:
        str: normalized name of register
    """
    return (GENERAL_REGISTER_STR + str(X86_REGISTER_SIZES[state.token.lower()])) if RE_GENERAL_REGISTER_MATCH.fullmatch(state.token) is not None \
        else RE_REMOVE_REGISTER_NUMBER.sub('', state.token)


def x86_memsize_value(self, state):
    """Replaces memory size pointers with 'memsize' followed by the value of that memsize in bytes

    Args:
        token (str): the current string token

    Returns:
        str: normalized memory size string
    """
    vsize, mem_str = MEM_SIZE_RE.fullmatch(state.token).groups()
    mem_size = X86_MEMORY_SIZES[mem_str.lower()] * (1 if vsize is None else int(vsize))
    return MEM_SIZE_TOKEN_STR + str(mem_size)



"""Atomically update tokens"""
import os
import pickle
import warnings
import time
import datetime
import numpy as np
import bincfg
from .misc_utils import get_module


# Whether or not we warn about atomic data not being able to be loaded when unpickling
_WARN_ATOMIC_DATA = True

def _set_warn_atomic_data(val):
    global _WARN_ATOMIC_DATA
    _WARN_ATOMIC_DATA = val

_ATOMIC_READ_RAISE_ERR = object()

# The number of seconds to wait before attempting to aquire a lock after failing
AQUIRE_LOCK_FAIL_WAIT_TIME_SECONDS = 0.1


class AtomicData:
    """A class that allows for atomic reading/updating of the given data to a pickle file
    
    Parameters
    ----------
        init_data: `Any`
            Data to initialize the atomic file with. If the atomic file already exists, then that data will be loaded
        filepath: `Optional[str]`
            An optional filepath to store the dictionary, otherwise will be stored at './atomic_dict.pkl'
        lockpath: `Optional[str]`
            An optional filepath for the lock file to use to atomically update the dictionary, otherwise will be
                stored at './.[filepath].lock' where [filepath] is the given `filepath` parameter
        max_read_attempts: `Optional[int]`
            An optional integer specifying the maximum number of attempts to atomically read this dictionary before
                giving up and raising an error. Set to None to attempt indefinitely. Defaults to None
        delete_file: `bool`
            If True, then the file and lockfile will be deleted on initialization to start from scratch
    """

    def __init__(self, init_data, filepath=None, lockpath=None, max_read_attempts=None, delete_file=False):
        self._filepath = './atomic_data.pkl' if filepath is None else filepath
        self._lock_path = os.path.join(os.path.dirname(self._filepath), '.%s.lock' % os.path.basename(self._filepath)) if lockpath is None else lockpath
        self._lock = None

        if max_read_attempts is not None and max_read_attempts <= 0:
            raise ValueError("max_read_attempts must be > 0: %d" % max_read_attempts)
        self._max_read_attempts = 2**100 if max_read_attempts is None else max_read_attempts

        # Delete the files if starting from scratch
        if delete_file:
            self.delete_file(force=True)

        # Get the initial data
        self.atomic_read(default=init_data)
    
    def atomic_read(self, default=_ATOMIC_READ_RAISE_ERR):
        """Atomically reads the data from file, updating self.data
        
        Args:
            default (Optional[Any]): If this is passed and the file does not already exist, then this data will be saved
                to file and set to self.data
        """
        with _AquireLock(self._max_read_attempts, self._lock_path):

            # If the path doesn't exist, check if we need to raise an error, or update the file
            if not os.path.exists(self._filepath):
                if default is not _ATOMIC_READ_RAISE_ERR:
                    self.data = default
                    self._locked_write()
                else:
                    raise FileNotFoundError('Could not find inital atomic file to read from, and `default` data was not passed: %s' % self._filepath)
            
            # Otherwise it does exist, update self
            else:
                self.data = self._locked_read()
        
        return self.data
    
    def _locked_read(self):
        """Reads the data from file, assuming a lock has already been aquired"""
        with open(self._filepath, 'rb') as f:
            return pickle.load(f)
    
    def atomic_write(self):
        """Atomically writes the data at self.data to the pickle file"""
        with _AquireLock(self._max_read_attempts, self._lock_path):
            self._locked_write()
    
    def _locked_write(self):
        """Writes the data at self.data to file, assuming a lock has already been aquired"""
        with open(self._filepath, 'wb') as f:
            pickle.dump(self.data, f)
    
    def atomic_update(self, update_func, *update_args, **update_kwargs):
        """Atomically updates the data
        
        Will first aquire a lock on the data, read it in, then call `update_func(file_data, update_data)` where `file_data`
        is the data from the current atomic file, then write the data back to file and finally release the lock.

        NOTE: this will prevent any and all updates to the atomic file until update_func has completed

        NOTE: any errors within the update_func will be handled properly and will likely not mess up the atomic file

        Args:
            update_func (Callable): function that takes in: the data currently saved in file, the current data, then the 
                passed args and kwargs, and returns the updated data to write back to file
            update_args (Any): args to pass to update_func, after the current data saved in file
            update_kwargs (Any): kwargs to pass to update_func
        
        Returns:
            Any: the updated data
        """
        with _AquireLock(self._max_read_attempts, self._lock_path):
            self.data = update_func(self._locked_read(), self.data, *update_args, **update_kwargs)
            self._locked_write()
            return self.data
    
    def aquire_lock(self):
        """Aquires the lock needed to update data
        
        NOTE: this will prevent any and all updates to the atomic file until self.release_lock() is called. Make sure
        you call it quickly or other processes may hang!

        NOTE: if the lock has already been aquired, nothing will happen

        NOTE: it can be dangerous to attempt to aquire locks yourself, as any errors raised must be handled nicely and
        self.release_lock() must be called otherwise other processes may hang
        """
        if self._lock is None:
            self._lock = _AquireLock(self._max_read_attempts, self._lock_path).__enter__()
    
    def release_lock(self):
        """Releases the lock. Assumes it has already been aquired, otherwise an error will be raised"""
        if self._lock is None:
            raise ValueError("release_lock() was called, but the lock has not been aquired!")
        self._lock.__exit__()
        self._lock = None
    
    def delete_file(self, force=False):
        """Atomically deletes the file being used"""
        if force:
            if os.path.exists(self._filepath):
                os.remove(self._filepath)
            if os.path.exists(self._lock_path):
                os.remove(self._lock_path)
        else:
            with _AquireLock(self._max_read_attempts, self._lock_path):
                if os.path.exists(self._filepath):
                    os.remove(self._filepath)
    
    def __len__(self):
        """Gives length of current self.data"""
        return len(self.data)
    
    def __getstate__(self):
        """Doesn't send the actual data itself, that will be loaded"""
        ret = self.__dict__.copy()
        #del ret['data']
        return ret

    def __setstate__(self, state):
        """Set the state as normal, but read in the data when done"""
        for k, v in state.items():
            setattr(self, k, v)
        try:
            self._max_read_attempts, old = 50, self._max_read_attempts
            self.atomic_read()
            self._max_read_attempts = old
        except Exception as e:
            if _WARN_ATOMIC_DATA:
                warnings.warn("Could not load atomic data from file: %s, due to %s: %s. This data could be outdated!" % (self._filepath, type(e).__name__, e))


class _AquireLock:
    """Context manager to aquire a file lock, and remove it when done"""
    def __init__(self, max_attempts, lock_path):
        self._max_attempts, self._lock_path = max_attempts, lock_path
        get_module('atomicwrites', err_message='Package is required for atomic dictionary file!')
        from atomicwrites import atomic_write
        self.atomic_write = atomic_write
    
    def __enter__(self):
        rng = np.random.default_rng(seed=os.getpid() * int(time.time() * 1_000_000))
        for i in range(self._max_attempts):
            try:
                with self.atomic_write(self._lock_path, overwrite=False) as f:
                    f.write(datetime.datetime.now().isoformat())
                time.sleep(rng.random() * 0.2)  # Wait on average 0.1 seconds before trying again

                return self
            except FileExistsError:
                pass
            
            time.sleep(AQUIRE_LOCK_FAIL_WAIT_TIME_SECONDS)

        raise AquireLockError(self._max_attempts, self._lock_path)
            
    def __exit__(self, exc_type, exc_value, exc_tb):
        if os.path.exists(self._lock_path):
            os.remove(self._lock_path)


class AquireLockError(Exception):
    def __init__(self, attempts, lock_path):
        super().__init__("Could not aquire file lock from file after %d attempts using lock path: %s" % (attempts, lock_path))


class AtomicTokenDict:
    """Acts like a normal token dictionary, but allows for atomic operations
    
    Parameters
    ----------
        init_data: `Optional[Dict[str, int]]`
            Data to initialize the atomic token dict with. If the atomic file already exists, then that data will be loaded
        filepath: `Optional[str]`
            An optional filepath to store the dictionary, otherwise will be stored at './atomic_dict.pkl'
        lockpath: `Optional[str]`
            An optional filepath for the lock file to use to atomically update the dictionary, otherwise will be
                stored at './.[filepath].lock' where [filepath] is the given `filepath` parameter
        max_read_attempts: `Optional[int]`
            An optional integer specifying the maximum number of attempts to atomically read this dictionary before
                giving up and raising an error. Set to None to attempt indefinitely. Defaults to None
        delete_file: `bool`
            If True, then the file and lockfile will be deleted on initialization to start from scratch
    """

    def __init__(self, init_data=None, filepath=None, lockpath=None, max_read_attempts=None, delete_file=False):
        self._data = AtomicData(init_data={}, filepath=filepath, lockpath=lockpath, 
                                max_read_attempts=max_read_attempts, delete_file=delete_file)
        
        # Check to make sure init_data is a valid type, and there are no duplicate tokens
        if init_data is not None:
            if not isinstance(init_data, (dict, AtomicTokenDict)):
                raise TypeError("Can only initialize AtomicTokenDict with data of type 'dict' or 'AtomicTokenDict', not %s" 
                                % repr(type(init_data).__name__))
            
            found = {}
            for k, v in init_data.items():
                if v in found:
                    raise ValueError("Found tokens with duplicate values in init_data: %s" % [(found[v], v), (k, v)])
                if k in self and self[k] != v:
                    raise ValueError("Found token %s in init_data and in loaded atomic data with different values: %d != %d" % (repr(k), v, self[k]))
                found[v] = k
            
            self.update(init_data)
    
    def __getitem__(self, key):
        return self.data[key]
    
    def __setitem__(self, key, value):
        if key in self.data:
            if self.data[key] != value:
                raise ValueError("Cannot set token key to a new value! key: %s, value: %s" % (repr(key), value))
            return
        
        self._atomic_update({key: value})
    
    def __contains__(self, key):
        return key in self.data
    
    def __len__(self):
        return len(self.data)
    
    def __str__(self):
        return repr(self)
    
    def __repr__(self):
        return repr(self.data)
    
    def __iter__(self):
        return iter(self.data)
    
    def update(self, tokens):
        """Updates this dictionary with the given tokens
        
        Args:
            tokens (Union[Dict[str, int], AtomicTokenDict]): dictionary mapping token strings to their integer values. 
                Any tokens in the dictionary that are not in this dictionary will be added, and any tokens that already 
                exist and have the same value will be ignored. If there are any tokens that already exist, but have a 
                different value, then an error will be raised
        """
        update_tokens = {}
        for k, v in tokens.items():
            if k in self.data:
                if self.data[k] != v:
                    raise ValueError("Cannot set token key to a new value! key: %s, value: %s" % (repr(k), v))
                continue
            update_tokens[k] = v
        
        if len(update_tokens) > 0:
            self._atomic_update(update_tokens)
    
    def items(self):
        return self.data.items()
    
    def values(self):
        return self.data.values()
    
    def keys(self):
        return self.data.keys()
    
    def get(self, key, default=None):
        return self.data.get(key, default=default)

    def setdefault(self, key, default=None):
        """If the key exists, return the value. Otherwise set the key to the given default (or len(self) if default=None)"""
        if key in self.data:
            return self.data[key]
        
        default = len(self) if default is None else default
        self._atomic_update({key: default})
        return default
    
    def addtokens(self, *tokens):
        """Adds the given tokens to this dictionary, ignoring any that already exist
        
        Args:
            tokens (str): arbitrary number of string tokens to add to this token dict
        """
        update_tokens = {}
        for t in tokens:
            if not isinstance(t, str):
                raise TypeError("Each token must be a string, not %s" % repr(type(t).__name__))
            if t not in self:
                update_tokens[t] = len(self) + len(update_tokens)
        
        if len(update_tokens) > 0:
            self._atomic_update(update_tokens)
    
    def _atomic_update(self, token_dict=None):
        """Atomically update the tokens from the given token_dict. Does no checks beforehand to see if there are any
        conflicts, duplicates, etc.
        
        Args:
            token_dict (Optional[Dict[str, int]]): token dictionary to update with, or None to just read in any updated
                tokens from file
        """
        self._data.atomic_update(bincfg.update_atomic_tokens, token_dict if token_dict is not None else {})
    
    def delete_file(self):
        "Deletes the atomic token dictinoary file"
        self._data.delete_file()

    @property
    def data(self):
        """Returns the token dictionary"""
        return self._data.data
    
    @property
    def inverse(self):
        """Return a new dict containing an inverse mapping of this current dictionary"""
        return {v: k for k, v in self.items()}
    
    @property
    def filepath(self):
        """Return the filepath being used to store the atomic data"""
        return self._data._filepath
    
    @property
    def lock_path(self):
        """Return the lock path being used to store the atomic data"""
        return self._data._lock_path
    
    def __hash__(self):
        import bincfg
        return bincfg.hash_obj(self.data)



import os
import pickle
from collections import Counter
from .cfg import CFG
from ..normalization import normalize_cfg_data, get_normalizer
from ..utils import progressbar, isinstance_with_iterables, hash_obj, eq_obj


class CFGDataset:
    """A dataset of ``CFG``'s.

    Parameters
    ----------
    cfg_data: `Optional[Union[CFG, CFGDataset, Iterable]]`
        a ``CFG``, ``CFGDataset`` or iterable of ``CFG``'s or ``CFGDataset``'s to add to this dataset, or None to
        initialize this ``CFGDataset`` empty
    normalizer: `Optional[Union[str, Normalizer]]`
        if not None, then a normalizer to use. Will normalize all incoming ``CFG``'s if they do not already have the name 
        normalization (will attempt to renormalize incoming ``CFG``'s if they already have a normalization). Can be a 
        ``Normalizer`` object or string.
    load_path: `str`
        if not None, loads all files in this directory that end with '.txt' or '.dot'. Will raise an error if there are 
        no files. Will ignore any files that end with '.txt' or '.dot', but cannot be parsed.
    max_files: `Optional[int]`
        stops after loading this many files. If None, then there is no max
    allow_multiple_norms: `bool`
        by default, ``CFGDataset`` will only allow unnormalized cfg's when `normalizer=None` (if `normalizer` is not None, 
        then any normalized cfg added will be renormalized). Setting `allow_multiple_norms` to True will allow this 
        ``CFGDataset`` to store cfg data with any normalization method (assuming `normalizer=None`)
    progress: `bool`
        if True, will show a progressbar when loading cfg's from load_path
    metadata: `Optional[Dict]`
        a dictionary of metadata to attach to this CFGDataset
        NOTE: passed dictionary will be shallow copied
    num_workers: `int`
        if > 1, then the loading of data using the `load_path` parameter will be split over this many processes
    add_data_kwargs: `Any`
        extra kwargs to pass to add_data while adding cfgs
    """

    cfgs = None
    """The list of all cfgs in this dataset"""

    normalizer = None
    """The normalizer used in this dataset, or None if there is no normalizer"""

    metadata = None
    """A dictionary of metadata associated with this ``CFGDataset``"""

    def __init__(self, cfg_data=None, normalizer=None, load_path=None, max_files=None, allow_multiple_norms=False, 
        progress=False, metadata=None, num_workers=1, **add_data_kwargs):

        self.allow_multiple_norms = allow_multiple_norms
        self.normalizer = get_normalizer(normalizer) if normalizer is not None else None
        self.metadata = {} if metadata is None else metadata.copy()
        self.cfgs = []

        if cfg_data is not None:
            self.add_data(*cfg_data, progress=progress, **add_data_kwargs)

        # Load in files if needed
        if load_path is not None:
            files = list(sorted([f for f in os.listdir(load_path) if f.endswith('.txt') or f.endswith('.dot')]))
            if len(files) == 0:
                raise ValueError("No files found ending in '.txt' or '.dot'")
            if max_files is not None:
                files = files[:max_files]
            
            if num_workers <= 1:
                for file in progressbar(files, progress=progress):
                    metadata = {'uid': file}
                    self.add_data(CFG(os.path.join(load_path, file), metadata=metadata), progress=False, **add_data_kwargs)
            else:
                if progress: print("MP loading cfgs...")
                res = get_thread_pool(num_workers=num_workers).map(_mp_load_cfg, [os.path.join(load_path, file) for file in files])
                if progress: print("Loading complete! Adding loaded cfgs...")
                self.add_data(*res, progress=progress, **add_data_kwargs)
    
    def add_data(self, *cfg_data, inplace=True, force_renormalize=False, progress=False):
        """Adds data to this dataset

        Args:
            cfg_data (Union[CFG, CFGDataset, Iterable]): arbitrary amount of ``CFG``/``CFGDataset``'s, or iterables of 
                them, to add to this dataset
            inplace (bool, optional): whether or not to normalize the incoming cfg_data inplace. Defaults to True.
            force_renormalize (bool, optional): by default, this method will only normalize cfg's whose 
                .normalizer != to this dataset's normalizer. However if `force_renormalize=True`, then all cfg's will 
                be renormalized even if they have been previously normalized with the same normalizer. Defaults to False.
            progress (bool, optional): if True, will show a progressbar when adding multiple cfgs. Defaults to False.

        Raises:
            TypeError: when attempting to add something that is not a ``CFG``, ``CFGDataset``, or iterables of them
            ValueError: when attempting to use multiple different normalizers and `self.allow_multiple_norms=False`
        """
        # Check that all elements in cfg_data are CFG's or CFGDataset's, or iterables of them
        temp = []
        for cfg in cfg_data:
            try:
                temp += isinstance_with_iterables(cfg, (CFG, CFGDataset), recursive=False, ret_list=True)
            except:
                raise TypeError("Can only add CFG's/CFGDataset's, or iterables of them, to CFGDataset, not '%s'" % type(cfg).__name__)
        cfg_data = temp

        # Check to see if we need to normalize
        if self.normalizer is not None:
            cfg_data = normalize_cfg_data(cfg_data, normalizer=self.normalizer, inplace=inplace, 
                force_renormalize=force_renormalize, convert_to_mem=False, unpack_cfgs=True, progress=progress)
        
        # Otherwise make sure normalizers are all None on data, and unpack cfgs
        else:
            temp_data = []
            for cfg in cfg_data:
                if cfg.normalizer is not None and not self.allow_multiple_norms:
                    raise ValueError("Found normalization '%s' on data to add to CFGDataset that has no normalizer" %
                        cfg.normalizer)
                temp_data += [cfg] if isinstance(cfg, CFG) else cfg.cfgs if isinstance(cfg, CFGDataset) else list(cfg)
            cfg_data = temp_data

        self.cfgs += cfg_data
    
    def normalize(self, normalizer=None, inplace=True, force_renormalize=False, progress=False):
        """Normalize this ``CFGDataset``.

        Args:
            normalizer (Union[str, Normalizer]): the normalizer to use. Can be a ``Normalizer`` object, or a 
                string, or None to use the default BaseNormalizer(). Defaults to None.
            inplace (bool, optional): by default, normalizes this dataset inplace (IE: without copying objects). Can set
                to False to return a copy. Defaults to True.
            force_renormalize (bool, optional): by default, this method will only normalize cfg's whose 
                .normalizer != to the passed normalizer. However if `force_renormalize=True`, then all cfg's will be 
                renormalized even if they have been previously normalized with the same normalizer.. Defaults to False.
            progress (bool, optional): if True, will show a progressbar while normalizing. Defaults to False.

        Returns:
            CFGDataset: this dataset normalized
        """
        return normalize_cfg_data(self, normalizer, inplace=inplace, force_renormalize=force_renormalize, progress=progress)
    
    @property
    def num_blocks(self):
        """Return total number of blocks across all cfg's"""
        return sum(cfg.num_blocks for cfg in self.cfgs)

    @property
    def num_functions(self):
        """Return total number of functions across all cfg's"""
        return sum(cfg.num_functions for cfg in self.cfgs)
        
    @property
    def num_edges(self):
        """Return total number of edges across all cfg's"""
        return sum(cfg.num_edges for cfg in self.cfgs)
        
    @property
    def num_asm_lines(self):
        """Return total number of assembly lines across all cfg's"""
        return sum(cfg.num_asm_lines for cfg in self.cfgs)
    
    @property
    def num_cfgs(self):
        """Return the number of cfgs in this dataset"""
        return len(self.cfgs)
    
    @property
    def asm_counts(self):
        """A collections.Counter() of all unique assembly lines and their counts accross all cfg's in this dataset"""
        return sum((cfg.asm_counts for cfg in self.cfgs), Counter())

    def __str__(self):
        stat_names = ["CFG's", 'Functions', 'Edges', 'Basic Blocks', 'Assembly Lines']
        c = [self.num_cfgs, self.num_functions, self.num_edges, self.num_blocks, self.num_asm_lines]
        stats = _get_stats(stat_names, c)
        norm = ('with normalizer: %s' % self.normalizer) if self.normalizer is not None else 'with no normalizer'
        return "%s %s\nStats:\n%s" % (self.__class__.__name__, norm, stats)

    def __repr__(self):
        return self.__str__()
    
    def __len__(self):
        return len(self.cfgs)
    
    def __getitem__(self, idx):
        return self.cfgs[idx]
    
    def __iter__(self):
        return iter(self.cfgs)
    
    def save(self, path):
        """Saves this CFGDataset to path"""
        with open(path, 'wb') as f:
            pickle.dump(self, f)
    
    def dumps(self):
        """Returns this object pickled with pickle.dumps()"""
        return pickle.dumps(self)
    
    @classmethod
    def load(cls, path):
        """Loads this CFGDataset from path"""
        with open(path, 'rb') as f:
            return pickle.load(f)
    
    def __hash__(self):
        return sum(hash(c) for c in self.cfgs) * 17 + hash_obj(self.metadata, return_int=True) * 31
    
    def __eq__(self, other):
        return isinstance(other, CFGDataset) and all(eq_obj(self, other, selector=s) for s in ['normalizer', 'metadata']) \
            and eq_obj(self, other, selector='cfgs')


def _get_stats(stat_names, counts):
    """Returns a nicely-printable set of statistics for the CFGDataset

    Args:
        stat_names (Iterable[str]): the names for each statistic
        counts (Iterable[int]): the values for each statistic
    
    Returns:
        str: the nicely formatted set of statistics for the CFGDataset
    """
    return '\n'.join([('\t' + name + ': ' + str(c)) for name, c in zip(stat_names, counts)])


def _mp_load_cfg(path):
    """Loads a cfg from the given path and returns it"""
    return CFG(path, metadata={'uid': path})


"""
Utilities for CFG/MemCFG objects and their datasets
"""

import numpy as np
import bincfg
from .type_utils import *
from .misc_utils import get_smallest_np_dtype


def get_address(obj: 'AddressLike') -> 'int':
    """Gets the integer address from the given object

    Args:
        obj (Union[str, int, Addressable]): a string, int, or object with a string/int `.address` attribute (should 
            always be positive)

    Raises:
        TypeError: `obj` is an unknown type
        ValueError: given address is negative

    Returns:
        int: the integer address
    """
    if isinstance(obj, str):
        ret = int(obj, 0)
    elif isinstance(obj, (int, np.integer)):
        ret = obj
    elif hasattr(obj, 'address'):
        ret = get_address(obj.address)
    else:
        raise TypeError("Cannot get address value from object of type: '%s'" % type(obj).__name__)

    if ret < 0:
        raise ValueError("Cannot have an address that is negative: %d" % obj)
    return ret


def get_special_function_names():
    """Returns the current global special function names"""
    return SPECIAL_FUNCTION_NAMES


def check_for_normalizer(dataset, cfg_data):
    """Checks the incoming data for a normalizer to set to be `dataset`'s normalizer
    
    Assumes this dataset does not yet have a normalizer. Searches the incoming `cfg_data` for a cfg/dataset that
    has a normalizer, and sets it to be this dataset's normalizer. If this method finds no normalizer, or multiple
    unique normalizers, then an error will be raised.

    Args:
        dataset (Union[CFGDataset, MemCFGDataset]): a ``CFGDataset`` or ``MemCFGDataset`` without a normalizer
        cfg_data (Iterable[Union[str, CFG, MemCFG, CFGDataset, MemCFGDataset]]): an iterable of 
            str/``CFG``/``MemCFG``/``CFGDataset``/``MemCFGDataset``'s

    Raises:
        ValueError: when there are multiple conflicting normalizers, or if no normalizer could be found
    """
    # If this is an individual object whose normalizer is the same for itself or all sub-objects, then we only need to check itself
    if isinstance(cfg_data, (bincfg.MemCFG, bincfg.CFG, bincfg.MemCFGDataset)) or \
        (isinstance(cfg_data, bincfg.CFGDataset) and not cfg_data.allow_multiple_norms):
        cfg_data = [cfg_data]
    
    for cfgd in cfg_data:
        if isinstance(cfgd, str):
            continue
        if cfgd.normalizer is not None:
            if dataset.normalizer is None:
                dataset.normalizer = cfgd.normalizer
            elif dataset.normalizer != cfgd.normalizer:
                raise ValueError("Multiple normalizers detected.")
    
    if dataset.normalizer is None:
        raise ValueError("Could not find a normalizer in cfg_data.")


def update_atomic_tokens(file_tokens, curr_data, update_tokens):
    """Updates atomic tokens. Only meant to be passed to AtomicData.atomic_update as the function to use"""
    curr_data.update(file_tokens)
    for t in update_tokens:
        curr_data.setdefault(t, len(curr_data))
    return curr_data


def update_memcfg_tokens(cfg_data, tokens):
    """Adds all new tokens to `tokens`, and updates all tokens in `cfg_data` to their respective values in `tokens`

    Tokens in `cfg_data` will be modified, as will the `.asm_lines` attribute of each memcfg. Assumes the `cfg_data`
    has conflicting tokens to `tokens` and thus needs modification. Both `cfg_data` and `tokens` will be 
    modified in-place.
    
    Args:
        cfg_data (Union[MemCFG, MemCFGDataset]): the memcfg/memcfgdataset to have its tokens changed
        tokens (Union[Dict[str, int], AtomicData]): the dictionary of tokens to update with the new tokens in `cfg_data`.
            Can be an AtomicData object for atomic updating of tokens
    """
    # If the tokens are the same, we can just return
    if tokens is cfg_data.tokens:
        return
        
    # Add all the new tokens in cfg_data to tokens
    if isinstance(tokens, bincfg.AtomicTokenDict):
        tokens.update(cfg_data.tokens)

    else:
        for new_token in cfg_data.tokens:
            tokens.setdefault(new_token, len(tokens))
    
    # Create the mapping from old token value to new token value
    old_to_new = {v: tokens[t] for t, v in cfg_data.tokens.items()}

    # Update all of the `.asm_lines` to their new token values
    update_cfgs = [cfg_data] if isinstance(cfg_data, bincfg.MemCFG) else cfg_data.cfgs
    for cfg in update_cfgs:
        new_asm_lines = [old_to_new[l] for l in cfg.asm_lines]
        cfg.asm_lines = np.array(new_asm_lines, dtype=get_smallest_np_dtype(max(new_asm_lines)))


# The default global set of special function names
SPECIAL_FUNCTION_NAMES = {'free', 'printf', 'fputc', 'fprintf', '__gmon_start__', 'memcpy', 'exit', 'fwrite', 'abort', 
'strcpy', 'memmove', '__errno_location', 'ferror', 'fread', 'fclose', 'sprintf', '__libc_start_main', 'ftell', 'fseek', 
'fopen', 'libintl_gettext', 'getenv', 'libintl_textdomain', 'strrchr', 'fputs', 'libintl_bindtextdomain', 'strerror', 
'_IO_putc', 'setlocale', 'vfprintf', 'xmalloc', '__xstat', 'xexit', 'unlink', 'getopt_long', 'close', 'fflush', 'strcmp',
'malloc', 'bfd_scan_vma', 'xrealloc', 'bfd_get_error', 'bfd_errmsg', 'strlen', 'bfd_set_format', 'strtol', 'access', 
'xstrdup', 'mkstemps', 'bfd_malloc', 'bfd_target_list', 'xmalloc_set_program_name', 'bfd_openw', '__assert_fail', 
'bfd_close_all_done', 'bfd_set_default_target', 'mkstemp', 'memset', 'bfd_arch_list', 'bfd_openr', 'ctime', 'strchr', 
'bfd_printable_arch_mach', 'realloc', 'mkdtemp', 'bfd_close', 'bfd_init', 'bfd_check_format_matches', 'bfd_check_format', 
'strncmp', '_IO_getc', 'puts', 'strncpy', 'putchar', 'bfd_set_error_program_name', 'bfd_get_section_by_name', 
'bfd_map_over_sections', 'strtoul', 'qsort', 'bfd_get_section_contents', '__strdup', 'filename_cmp', 'open', 'write', 
'strstr', 'bfd_openr_next_archived_file', 'bfd_set_section_contents', 'bfd_get_arch', 'bfd_set_section_size', 'strcat', 
'sbrk', 'memcmp', 'bfd_get_mach', 'bfd_canonicalize_reloc', 'bfd_get_reloc_upper_bound', 'read', 'concat', 'xcalloc', 
'feof', 'stpcpy', 'bfd_set_error', 'lbasename', 'bfd_set_symtab', 'bfd_make_section_with_flags', 'strcasecmp', 
'unlink_if_ordinary', 'bfd_bread', 'bfd_seek', 'snprintf', 'calloc', 'cplus_demangle_name_to_style', 
'cplus_demangle_set_style', 'fileno', 'perror', 'bfd_reloc_type_lookup', 'xstrerror', '_exit', 'fcntl', 'fdopen', 
'__lxstat', 'chmod', 'bsearch', 'fnmatch', 'time', 'bfd_demangle', 'bfd_set_section_flags', 'cplus_demangle', 
'remove', 'dup2', 'pipe', 'kill', 'wait4', 'waitpid', 'vfork', 'execvp', 'sleep', 'execv', 'bfd_hash_traverse', 
'bfd_hash_newfunc', 'bfd_hash_table_free', 'bfd_hash_allocate', 'bfd_hash_lookup', 'bfd_scan_arch', 'bfd_hash_table_init', 
'bfd_set_start_address', 'atoi', 'rewind', 'htab_find_slot', 'htab_find', 'abort', 'abs', 'acos', 'asctime', 'asctime_r', 
'asin', 'assert', 'atan', 'atan2', 'atexit', 'atof', 'atoi', 'atol', 'bsearch', 'btowc', 'calloc', 'catclose', 'catgets', 
'catopen', 'ceil', 'clearerr', 'clock', 'cos', 'cosh', 'ctime', 'ctime64', 'ctime_r', 'ctime64_r', 'difftime', 
'difftime64', 'div', 'erf', 'erfc', 'exit', 'exp', 'fabs', 'fclose', 'fdopen', 'feof', 'ferror', 'fflush', 'fgetc', 
'fgetpos', 'fgets', 'fgetwc', 'fgetws', 'fileno', 'floor', 'fmod', 'fopen', 'fprintf', 'fputc', 'fputs', 'fputwc', 
'fputws', 'fread', 'free', 'freopen', 'frexp', 'fscanf', 'fseek', 'fsetpos', 'ftell', 'fwide', 'fwprintf', 'fwrite', 
'fwscanf', 'gamma', 'getc', 'getchar', 'getenv', 'gets', 'getwc', 'getwchar', 'gmtime', 'gmtime64', 'gmtime_r', 
'gmtime64_r', 'hypot', 'isalnum', 'isalpha', 'isascii', 'isblank', 'iscntrl', 'isdigit', 'isgraph', 'islower', 
'isprint', 'ispunct', 'isspace', 'isupper', 'iswalnum', 'iswalpha', 'iswblank', 'iswcntrl', 'iswctype', 'iswdigit', 
'iswgraph', 'iswlower', 'iswprint', 'iswpunct', 'iswspace', 'iswupper', 'iswxdigit', 'isxdigit', 'j0', 'j1', 'jn', 
'labs', 'ldexp', 'ldiv', 'localeconv', 'localtime', 'localtime64', 'localtime_r', 'localtime64_r', 'log', 'log10', 
'longjmp', 'malloc', 'mblen', 'mbrlen', 'mbrtowc', 'mbsinit', 'mbsrtowcs', 'mbstowcs', 'mbtowc', 'memchr', 'memcmp', 
'memcpy', 'memmove', 'memset', 'mktime', 'mktime64', 'modf', 'nextafter', 'nextafterl', 'nexttoward', 'nexttowardl', 
'nl_langinfo4', 'perror', 'pow', 'printf', 'putc', 'putchar', 'putenv', 'puts', 'putwc', 'putwchar', 'qsort', 
'quantexpd32', 'quantexpd64', 'quantexpd128', 'quantized32', 'quantized64', 'quantized128', 'samequantumd32', 
'samequantumd64', 'samequantumd128', 'raise', 'rand', 'rand_r', 'realloc', 'regcomp', 'regerror', 'regexec', 
'regfree', 'remove', 'rename', 'rewind', 'scanf', 'setbuf', 'setjmp', 'setlocale', 'setvbuf', 'signal', 'sin', 
'sinh', 'snprintf', 'sprintf', 'sqrt', 'srand', 'sscanf', 'strcasecmp', 'strcat', 'strchr', 'strcmp', 'strcoll', 
'strcpy', 'strcspn', 'strerror', 'strfmon', 'strftime', 'strlen', 'strncasecmp', 'strncat', 'strncmp', 'strncpy', 
'strpbrk', 'strptime', 'strrchr', 'strspn', 'strstr', 'strtod', 'strtod32', 'strtod64', 'strtod128', 'strtof', 
'strtok', 'strtok_r', 'strtol', 'strtold', 'strtoul', 'strxfrm', 'swprintf', 'swscanf', 'system', 'tan', 'tanh', 
'time', 'time64', 'tmpfile', 'tmpnam', 'toascii', 'tolower', 'toupper', 'towctrans', 'towlower', 'towupper', 
'ungetc', 'ungetwc', 'va_arg', 'va_copy', 'va_end', 'va_start', 'vfprintf', 'vfscanf', 'vfwprintf', 'vfwscanf', 
'vprintf', 'vscanf', 'vsprintf', 'vsnprintf', 'vsscanf', 'vswprintf', 'vswscanf', 'vwprintf', 'vwscanf', 'wcrtomb', 
'wcscat', 'wcschr', 'wcscmp', 'wcscoll', 'wcscpy', 'wcscspn', 'wcsftime', 'wcslen', 'wcslocaleconv', 'wcsncat', 
'wcsncmp', 'wcsncpy', 'wcspbrk', 'wcsptime', 'wcsrchr', 'wcsrtombs', 'wcsspn', 'wcsstr', 'wcstod', 'wcstod32', 
'wcstod64', 'wcstod128', 'wcstof', 'wcstok', 'wcstol', 'wcstold', 'wcstombs', 'wcstoul', 'wcsxfrm', 'wctob', 'wctomb', 
'wctrans', 'wctype', 'wcwidth', 'wmemchr', 'wmemcmp', 'wmemcpy', 'wmemmove', 'wmemset', 'wprintf', 'wscanf', 'y0', 'y1', 
'yn'}
